# Copyright contributors to the Terratorch project

import os
import sys
import pytest
import torch
import torch.distributed as dist
import lightning.pytorch as pl
from lightning.pytorch import Trainer
from typing import Any

# Check for dependencies
try:
    import prithviwxc
    import granitewxc
    from granitewxc.utils.config import get_config
    from granitewxc.utils.data import _get_transforms
    # Only import internal terratorch modules if deps are present 
    # or wrap them in the same check if they depend on wxc packages
    from terratorch.models.wxc_model_factory import WxCModelFactory
    from terratorch.tasks.wxc_task import WxCTask
    from terratorch.datamodules.era5 import ERA5DataModule
    from terratorch.datamodules.merra2_downscale import Merra2DownscaleNonGeoDataModule

    HAS_WXC_DEPS = True
except ImportError:
    HAS_WXC_DEPS = False



pytestmark = pytest.mark.skipif(
    not HAS_WXC_DEPS, 
    reason="PrithviWxC or GraniteWxC packages are not installed"
)

def setup_function():
    print("\nSetup function is called")

def teardown_function():
    try:
        os.remove("config.yaml")
    except OSError:
        pass



def test_wxc_unet_pincer_inference():
    try:
        os.environ['MASTER_ADDR'] = 'localhost'
        os.environ['MASTER_PORT'] = '12355'

        if dist.is_initialized():
            dist.destroy_process_group()

        dist.init_process_group(
            backend='gloo',
            init_method='env://',
            rank=0,
            world_size=1
        )

        hf_hub_download(
            repo_id="Prithvi-WxC/Gravity_wave_Parameterization",
            filename=f"magnet-flux-uvtp122-epoch-99-loss-0.1022.pt",
            local_dir=".",
        )

        hf_hub_download(
            repo_id="Prithvi-WxC/Gravity_wave_Parameterization",
            repo_type='dataset',
            filename=f"wxc_input_u_v_t_p_output_theta_uw_vw_era5_training_data_hourly_2015_constant_mu_sigma_scaling05.nc",
            local_dir=".",
        )

        model_args = {
            "in_channels": 1280,
            "input_size_time": 1,
            "n_lats_px": 64,
            "n_lons_px": 128,
            "patch_size_px": [2, 2],
            "mask_unit_size_px": [8, 16],
            "mask_ratio_inputs": 0.5,
            "embed_dim": 2560,
            "n_blocks_encoder": 12,
            "n_blocks_decoder": 2,
            "mlp_multiplier": 4,
            "n_heads": 16,
            "dropout": 0.0,
            "drop_path": 0.05,
            "parameter_dropout": 0.0,
            "residual": "none",
            "masking_mode": "both",
            "decoder_shifting": False,
            "positional_encoding": "absolute",
            "checkpoint_encoder": [3, 6, 9, 12, 15, 18, 21, 24],
            "checkpoint_decoder": [1, 3],
            "in_channels_static": 3,
            "input_scalers_mu": torch.tensor([0] * 1280),
            "input_scalers_sigma": torch.tensor([1] * 1280),
            "input_scalers_epsilon": 0,
            "static_input_scalers_mu": torch.tensor([0] * 3),
            "static_input_scalers_sigma": torch.tensor([1] * 3),
            "static_input_scalers_epsilon": 0,
            "output_scalers": torch.tensor([0] * 1280),
            "backbone_weights": "magnet-flux-uvtp122-epoch-99-loss-0.1022.pt",
            "backbone": "prithviwxc",
            "aux_decoders": "unetpincer",
        }
        task = WxCTask('WxCModelFactory', model_args=model_args, mode='eval')

        trainer = Trainer(
            max_epochs=1,
            limit_predict_batches=1,
        )
        dm = ERA5DataModule(train_data_path='.', valid_data_path='.')
        results = trainer.predict(model=task, datamodule=dm, return_predictions=True)

    except RuntimeError as e:
        print(e)
        dist.destroy_process_group()


def test_wxc_downscaling_pincer_instantiate():
    kwargs = {
        "in_channels": 1280,
        "input_size_time": 1,
        "n_lats_px": 64,
        "n_lons_px": 128,
        "in_channels_static": 3,
        "input_scalers_mu": torch.tensor([0] * 1280),
        "input_scalers_sigma": torch.tensor([1] * 1280),
        "input_scalers_epsilon": 0,
        "static_input_scalers_mu": torch.tensor([0] * 3),
        "static_input_scalers_sigma": torch.tensor([1] * 3),
        "static_input_scalers_epsilon": 0,
        "output_scalers": torch.tensor([0] * 1280),
        "patch_size_px": [2, 2],
        "mask_unit_size_px": [8, 16],
        "mask_ratio_inputs": 0.5,
        "embed_dim": 2560,
        "n_blocks_encoder": 12,
        "n_blocks_decoder": 2,
        "mlp_multiplier": 4,
        "n_heads": 16,
        "dropout": 0.0,
        "drop_path": 0.05,
        "parameter_dropout": 0.0,
        "residual": "none",
        "masking_mode": "both",
        "positional_encoding": "absolute",
        "config_path": "./tests/test_prithvi_wxc_model_factory_config.yaml",
        "wxc_auxiliary_data_path": ".",
        "checkpoint_path": "./pytorch_model.bin",
    }

    config_path = hf_hub_download(repo_id="ibm-granite/granite-geospatial-wxc-downscaling",
                                  filename="config.yaml",
                                  local_dir="./")
    config = get_config(config_path)
    config.download_path = './'
    hf_hub_download(repo_id="ibm-granite/granite-geospatial-wxc-downscaling", filename="pytorch_model.bin",
                    local_dir="./")

    hf_hub_download(repo_id="ibm-nasa-geospatial/Prithvi-WxC-1.0-2300M", filename="merra-2/MERRA2_sfc_20200101.nc",
                    local_dir=config.download_path)
    hf_hub_download(repo_id="ibm-nasa-geospatial/Prithvi-WxC-1.0-2300M", filename="merra-2/MERRA_pres_20200101.nc",
                    local_dir=config.download_path)
    hf_hub_download(repo_id="ibm-nasa-geospatial/Prithvi-WxC-1.0-2300M",
                    filename="climatology/anomaly_variance_surface.nc", local_dir=config.download_path)
    hf_hub_download(repo_id="ibm-nasa-geospatial/Prithvi-WxC-1.0-2300M",
                    filename="climatology/anomaly_variance_vertical.nc", local_dir=config.download_path)
    hf_hub_download(repo_id="ibm-nasa-geospatial/Prithvi-WxC-1.0-2300M", filename="climatology/musigma_surface.nc",
                    local_dir=config.download_path)
    hf_hub_download(repo_id="ibm-nasa-geospatial/Prithvi-WxC-1.0-2300M", filename="climatology/musigma_vertical.nc",
                    local_dir=config.download_path)
    hf_hub_download(repo_id="ibm-nasa-geospatial/Prithvi-WxC-1.0-2300M",
                    filename="climatology/climate_surface_doy001_hour00.nc", local_dir=config.download_path)
    hf_hub_download(repo_id="ibm-nasa-geospatial/Prithvi-WxC-1.0-2300M",
                    filename="climatology/climate_surface_doy001_hour03.nc", local_dir=config.download_path)
    hf_hub_download(repo_id="ibm-nasa-geospatial/Prithvi-WxC-1.0-2300M",
                    filename="climatology/climate_surface_doy001_hour06.nc", local_dir=config.download_path)
    hf_hub_download(repo_id="ibm-nasa-geospatial/Prithvi-WxC-1.0-2300M",
                    filename="climatology/climate_surface_doy001_hour09.nc", local_dir=config.download_path)
    hf_hub_download(repo_id="ibm-nasa-geospatial/Prithvi-WxC-1.0-2300M",
                    filename="climatology/climate_surface_doy001_hour12.nc", local_dir=config.download_path)
    hf_hub_download(repo_id="ibm-nasa-geospatial/Prithvi-WxC-1.0-2300M",
                    filename="climatology/climate_surface_doy001_hour15.nc", local_dir=config.download_path)
    hf_hub_download(repo_id="ibm-nasa-geospatial/Prithvi-WxC-1.0-2300M",
                    filename="climatology/climate_surface_doy001_hour18.nc", local_dir=config.download_path)
    hf_hub_download(repo_id="ibm-nasa-geospatial/Prithvi-WxC-1.0-2300M",
                    filename="climatology/climate_surface_doy001_hour21.nc", local_dir=config.download_path)
    hf_hub_download(repo_id="ibm-nasa-geospatial/Prithvi-WxC-1.0-2300M",
                    filename="climatology/climate_surface_doy001_hour00.nc", local_dir=config.download_path)
    hf_hub_download(repo_id="ibm-nasa-geospatial/Prithvi-WxC-1.0-2300M",
                    filename="climatology/climate_vertical_doy001_hour00.nc", local_dir=config.download_path)
    hf_hub_download(repo_id="ibm-nasa-geospatial/Prithvi-WxC-1.0-2300M",
                    filename="climatology/climate_vertical_doy001_hour03.nc", local_dir=config.download_path)
    hf_hub_download(repo_id="ibm-nasa-geospatial/Prithvi-WxC-1.0-2300M",
                    filename="climatology/climate_vertical_doy001_hour06.nc", local_dir=config.download_path)
    hf_hub_download(repo_id="ibm-nasa-geospatial/Prithvi-WxC-1.0-2300M",
                    filename="climatology/climate_vertical_doy001_hour09.nc", local_dir=config.download_path)
    hf_hub_download(repo_id="ibm-nasa-geospatial/Prithvi-WxC-1.0-2300M",
                    filename="climatology/climate_vertical_doy001_hour12.nc", local_dir=config.download_path)
    hf_hub_download(repo_id="ibm-nasa-geospatial/Prithvi-WxC-1.0-2300M",
                    filename="climatology/climate_vertical_doy001_hour15.nc", local_dir=config.download_path)
    hf_hub_download(repo_id="ibm-nasa-geospatial/Prithvi-WxC-1.0-2300M",
                    filename="climatology/climate_vertical_doy001_hour18.nc", local_dir=config.download_path)
    hf_hub_download(repo_id="ibm-nasa-geospatial/Prithvi-WxC-1.0-2300M",
                    filename="climatology/climate_vertical_doy001_hour21.nc", local_dir=config.download_path)
    hf_hub_download(repo_id="ibm-nasa-geospatial/Prithvi-WxC-1.0-2300M",
                    filename="climatology/climate_vertical_doy001_hour00.nc", local_dir=config.download_path)

    WxCModelFactory().build_model(backbone="prithviwxc", aux_decoders="downscaler", **kwargs)



def test_wxc_downscaling_pincer_task():
    model_args = {
        "backbone":  "prithviwxc",
        "aux_decoders": "downscaler",
        "in_channels": 1280,
        "input_size_time": 1,
        "n_lats_px": 64,
        "n_lons_px": 128,
        "in_channels_static": 3,
        "input_scalers_mu": torch.tensor([0] * 1280),
        "input_scalers_sigma": torch.tensor([1] * 1280),
        "input_scalers_epsilon": 0,
        "static_input_scalers_mu": torch.tensor([0] * 3),
        "static_input_scalers_sigma": torch.tensor([1] * 3),
        "static_input_scalers_epsilon": 0,
        "output_scalers": torch.tensor([0] * 1280),
        "patch_size_px": [2, 2],
        "mask_unit_size_px": [8, 16],
        "mask_ratio_inputs": 0.5,
        "embed_dim": 2560,
        "n_blocks_encoder": 12,
        "n_blocks_decoder": 2,
        "mlp_multiplier": 4,
        "n_heads": 16,
        "dropout": 0.0,
        "drop_path": 0.05,
        "parameter_dropout": 0.0,
        "residual": "none",
        "masking_mode": "both",
        "positional_encoding": "absolute",
        "config_path": "./tests/test_prithvi_wxc_model_factory_config.yaml",
        "wxc_auxiliary_data_path": ".",
        "checkpoint_path": "./pytorch_model.bin",
    }

    task = WxCTask('WxCModelFactory', model_args=model_args, mode='train')


def test_wxc_downscaling_pincer_predict():
    model_args = {
        "backbone":  "prithviwxc",
        "aux_decoders": "downscaler",
        "in_channels": 1280,
        "input_size_time": 1,
        "n_lats_px": 64,
        "n_lons_px": 128,
        "in_channels_static": 3,
        "input_scalers_mu": torch.tensor([0] * 1280),
        "input_scalers_sigma": torch.tensor([1] * 1280),
        "input_scalers_epsilon": 0,
        "static_input_scalers_mu": torch.tensor([0] * 3),
        "static_input_scalers_sigma": torch.tensor([1] * 3),
        "static_input_scalers_epsilon": 0,
        "output_scalers": torch.tensor([0] * 1280),
        "patch_size_px": [2, 2],
        "mask_unit_size_px": [8, 16],
        "mask_ratio_inputs": 0.5,
        "embed_dim": 2560,
        "n_blocks_encoder": 12,
        "n_blocks_decoder": 2,
        "mlp_multiplier": 4,
        "n_heads": 16,
        "dropout": 0.0,
        "drop_path": 0.05,
        "parameter_dropout": 0.0,
        "residual": "none",
        "masking_mode": "both",
        "positional_encoding": "absolute",
        "config_path": "./tests/test_prithvi_wxc_model_factory_config.yaml",
        "wxc_auxiliary_data_path": ".",
        "checkpoint_path": "./pytorch_model.bin",
    }


    task = WxCTask('WxCModelFactory', model_args=model_args, mode='train')

    config_path = hf_hub_download(repo_id="ibm-granite/granite-geospatial-wxc-downscaling",
                                  filename="config.yaml",
                                  local_dir="./")
    config = get_config(config_path)
    config.download_path = './'

    hf_hub_download(repo_id="ibm-granite/granite-geospatial-wxc-downscaling", filename="pytorch_model.bin",
                    local_dir="./")
    hf_hub_download(repo_id="ibm-nasa-geospatial/Prithvi-WxC-1.0-2300M", filename="merra-2/MERRA2_sfc_20200101.nc",
                    local_dir=config.download_path)
    hf_hub_download(repo_id="ibm-nasa-geospatial/Prithvi-WxC-1.0-2300M", filename="merra-2/MERRA_pres_20200101.nc",
                    local_dir=config.download_path)
    hf_hub_download(repo_id="ibm-nasa-geospatial/Prithvi-WxC-1.0-2300M",
                    filename="climatology/anomaly_variance_surface.nc", local_dir=config.download_path)
    hf_hub_download(repo_id="ibm-nasa-geospatial/Prithvi-WxC-1.0-2300M",
                    filename="climatology/anomaly_variance_vertical.nc", local_dir=config.download_path)
    hf_hub_download(repo_id="ibm-nasa-geospatial/Prithvi-WxC-1.0-2300M", filename="climatology/musigma_surface.nc",
                    local_dir=config.download_path)
    hf_hub_download(repo_id="ibm-nasa-geospatial/Prithvi-WxC-1.0-2300M", filename="climatology/musigma_vertical.nc",
                    local_dir=config.download_path)
    hf_hub_download(repo_id="ibm-nasa-geospatial/Prithvi-WxC-1.0-2300M",
                    filename="climatology/climate_surface_doy001_hour00.nc", local_dir=config.download_path)
    hf_hub_download(repo_id="ibm-nasa-geospatial/Prithvi-WxC-1.0-2300M",
                    filename="climatology/climate_surface_doy001_hour03.nc", local_dir=config.download_path)
    hf_hub_download(repo_id="ibm-nasa-geospatial/Prithvi-WxC-1.0-2300M",
                    filename="climatology/climate_surface_doy001_hour06.nc", local_dir=config.download_path)
    hf_hub_download(repo_id="ibm-nasa-geospatial/Prithvi-WxC-1.0-2300M",
                    filename="climatology/climate_surface_doy001_hour09.nc", local_dir=config.download_path)
    hf_hub_download(repo_id="ibm-nasa-geospatial/Prithvi-WxC-1.0-2300M",
                    filename="climatology/climate_surface_doy001_hour12.nc", local_dir=config.download_path)
    hf_hub_download(repo_id="ibm-nasa-geospatial/Prithvi-WxC-1.0-2300M",
                    filename="climatology/climate_surface_doy001_hour15.nc", local_dir=config.download_path)
    hf_hub_download(repo_id="ibm-nasa-geospatial/Prithvi-WxC-1.0-2300M",
                    filename="climatology/climate_surface_doy001_hour18.nc", local_dir=config.download_path)
    hf_hub_download(repo_id="ibm-nasa-geospatial/Prithvi-WxC-1.0-2300M",
                    filename="climatology/climate_surface_doy001_hour21.nc", local_dir=config.download_path)
    hf_hub_download(repo_id="ibm-nasa-geospatial/Prithvi-WxC-1.0-2300M",
                    filename="climatology/climate_surface_doy001_hour00.nc", local_dir=config.download_path)
    hf_hub_download(repo_id="ibm-nasa-geospatial/Prithvi-WxC-1.0-2300M",
                    filename="climatology/climate_vertical_doy001_hour00.nc", local_dir=config.download_path)
    hf_hub_download(repo_id="ibm-nasa-geospatial/Prithvi-WxC-1.0-2300M",
                    filename="climatology/climate_vertical_doy001_hour03.nc", local_dir=config.download_path)
    hf_hub_download(repo_id="ibm-nasa-geospatial/Prithvi-WxC-1.0-2300M",
                    filename="climatology/climate_vertical_doy001_hour06.nc", local_dir=config.download_path)
    hf_hub_download(repo_id="ibm-nasa-geospatial/Prithvi-WxC-1.0-2300M",
                    filename="climatology/climate_vertical_doy001_hour09.nc", local_dir=config.download_path)
    hf_hub_download(repo_id="ibm-nasa-geospatial/Prithvi-WxC-1.0-2300M",
                    filename="climatology/climate_vertical_doy001_hour12.nc", local_dir=config.download_path)
    hf_hub_download(repo_id="ibm-nasa-geospatial/Prithvi-WxC-1.0-2300M",
                    filename="climatology/climate_vertical_doy001_hour15.nc", local_dir=config.download_path)
    hf_hub_download(repo_id="ibm-nasa-geospatial/Prithvi-WxC-1.0-2300M",
                    filename="climatology/climate_vertical_doy001_hour18.nc", local_dir=config.download_path)
    hf_hub_download(repo_id="ibm-nasa-geospatial/Prithvi-WxC-1.0-2300M",
                    filename="climatology/climate_vertical_doy001_hour21.nc", local_dir=config.download_path)
    hf_hub_download(repo_id="ibm-nasa-geospatial/Prithvi-WxC-1.0-2300M",
                    filename="climatology/climate_vertical_doy001_hour00.nc", local_dir=config.download_path)

    config.data.data_path_surface = os.path.join(config.download_path, 'merra-2')
    config.data.data_path_vertical = os.path.join(config.download_path, 'merra-2')
    config.data.climatology_path_surface = os.path.join(config.download_path, 'climatology')
    config.data.climatology_path_vertical = os.path.join(config.download_path, 'climatology')

    config.model.input_scalers_surface_path = os.path.join(config.download_path, 'climatology/musigma_surface.nc')
    config.model.input_scalers_vertical_path = os.path.join(config.download_path, 'climatology/musigma_vertical.nc')
    config.model.output_scalers_surface_path = os.path.join(config.download_path,
                                                            'climatology/anomaly_variance_surface.nc')
    config.model.output_scalers_vertical_path = os.path.join(config.download_path,
                                                             'climatology/anomaly_variance_vertical.nc')

    dm = Merra2DownscaleNonGeoDataModule(
        time_range=('2020-01-01T00:00:00', '2020-01-01T23:59:59'),
        data_path_surface = config.data.data_path_surface,
        data_path_vertical = config.data.data_path_vertical,
        climatology_path_surface = config.data.climatology_path_surface,
        climatology_path_vertical = config.data.climatology_path_vertical,
        input_surface_vars = config.data.input_surface_vars,
        input_static_surface_vars = config.data.input_static_surface_vars,
        input_vertical_vars = config.data.input_vertical_vars,
        input_levels = config.data.input_levels,
        n_input_timestamps = config.data.n_input_timestamps,
        output_vars=config.data.output_vars,
        transforms=_get_transforms(config),
    )
    dm.setup('predict')

    trainer = Trainer(
        max_epochs=1,
        limit_predict_batches=1,
    )
    results = trainer.predict(model=task, datamodule=dm)

""" TODO deactivated because memory footprint too high
def test_wxc_unet_pincer_train():
    os.environ['MASTER_ADDR'] = 'localhost'
    os.environ['MASTER_PORT'] = '12355'

    if dist.is_initialized():
        dist.destroy_process_group()

    dist.init_process_group(
        backend='gloo',
        init_method='env://',
        rank=0,
        world_size=1
    )

    hf_hub_download(
        repo_id="Prithvi-WxC/Gravity_wave_Parameterization",
        filename=f"magnet-flux-uvtp122-epoch-99-loss-0.1022.pt",
        local_dir=".",
    )

    hf_hub_download(
        repo_id="Prithvi-WxC/Gravity_wave_Parameterization",
        filename=f"config.yaml",
        local_dir=".",
    )

    hf_hub_download(
        repo_id="Prithvi-WxC/Gravity_wave_Parameterization",
        repo_type='dataset',
        filename=f"wxc_input_u_v_t_p_output_theta_uw_vw_era5_training_data_hourly_2015_constant_mu_sigma_scaling05.nc",
        local_dir=".",
    )

    model_args = {
        "in_channels": 1280,
        "input_size_time": 1,
        "n_lats_px": 64,
        "n_lons_px": 128,
        "patch_size_px": [2, 2],
        "mask_unit_size_px": [8, 16],
        "mask_ratio_inputs": 0.5,
        "embed_dim": 2560,
        "n_blocks_encoder": 12,
        "n_blocks_decoder": 2,
        "mlp_multiplier": 4,
        "n_heads": 16,
        "dropout": 0.0,
        "drop_path": 0.05,
        "parameter_dropout": 0.0,
        "residual": "none",
        "masking_mode": "both",
        "decoder_shifting": False,
        "positional_encoding": "absolute",
        "checkpoint_encoder": [3, 6, 9, 12, 15, 18, 21, 24],
        "checkpoint_decoder": [1, 3],
        "in_channels_static": 3,
        "input_scalers_mu": torch.tensor([0] * 1280),
        "input_scalers_sigma": torch.tensor([1] * 1280),
        "input_scalers_epsilon": 0,
        "static_input_scalers_mu": torch.tensor([0] * 3),
        "static_input_scalers_sigma": torch.tensor([1] * 3),
        "static_input_scalers_epsilon": 0,
        "output_scalers": torch.tensor([0] * 1280),
        "backbone_weights": "magnet-flux-uvtp122-epoch-99-loss-0.1022.pt",
        "backbone": "prithviwxc",
        "aux_decoders": "unetpincer",
        "skip_connection": True,
    }

    task = WxCTask('WxCModelFactory', model_args=model_args, mode='train')

    trainer = Trainer(
        max_epochs=1,
        limit_predict_batches=1,
    )
    dm = ERA5DataModule(train_data_path='.', valid_data_path='.')
    results = trainer.fit(model=task, datamodule=dm)

    dist.destroy_process_group()
"""



# Copyright contributors to the Terratorch project

from torch import nn


class SegmentationHead(nn.Module):
    """Segmentation head"""

    def __init__(
        self, in_channels: int, num_classes: int, channel_list: list[int] | None = None, dropout: float = 0
    ) -> None:
        """Constructor

        Args:
            in_channels (int): Number of input channels
            num_classes (int): Number of output classes
            channel_list (list[int] | None, optional):  List with number of channels for each Conv
                layer to be created. Defaults to None.
            dropout (float, optional): Dropout value to apply. Defaults to 0.
        """
        super().__init__()
        self.num_classes = num_classes
        if channel_list is None:
            pre_head = nn.Identity()
        else:

            def block(in_channels, out_channels):
                return nn.Sequential(
                    nn.Conv2d(in_channels=in_channels, out_channels=out_channels, kernel_size=3, padding=1), nn.ReLU()
                )

            channel_list = [in_channels, *channel_list]
            pre_head = nn.Sequential(
                *[block(channel_list[i], channel_list[i + 1]) for i in range(len(channel_list) - 1)]
            )
            in_channels = channel_list[-1]
        dropout = nn.Identity() if dropout == 0 else nn.Dropout(dropout)
        self.head = nn.Sequential(
            pre_head,
            dropout,
            nn.Conv2d(
                in_channels=in_channels,
                out_channels=num_classes,
                kernel_size=1,
            ),
        )

    def forward(self, x):
        return self.head(x)



"""Max coverage tests for vit_adapter_modules.

Stubs heavy external dependencies (MSDeformAttn, SyncBatchNorm) to isolate logic.
Covers:
- get_reference_points
- deform_inputs
- SpatialPriorModule forward (with_cp True/False)
- DWConv token reshaping logic
- ConvFFN end-to-end transformation
- Extractor with and without CFFN + checkpoint flag branches
- Injector residual with gamma scaling + checkpoint branch
- InteractionBlock with extra_extractor on/off
"""
from __future__ import annotations

import sys
import types
import pytest
import torch
import torch.nn as nn

# ---- Stubs for external deps before import ----
# Stub MSDeformAttn with simple additive attention-like operation
if 'terratorch.models.backbones.detr_ops.modules' not in sys.modules:
    mods = types.ModuleType('terratorch.models.backbones.detr_ops.modules')
    class StubMSDeformAttn(nn.Module):
        def __init__(self, d_model: int, n_levels: int, n_heads: int, n_points: int, ratio: float):
            super().__init__()
            self.proj = nn.Linear(d_model, d_model)
        def forward(self, query, reference_points, feat, spatial_shapes, level_start_index, padding_mask=None):
            # ignore spatial args; simple fusion
            q = self.proj(query)
            # truncate / pad feat to query length
            feat = feat[:, : query.shape[1], :]
            return q + feat
    mods.MSDeformAttn = StubMSDeformAttn
    sys.modules['terratorch.models.backbones.detr_ops.modules'] = mods

# Stub SyncBatchNorm to standard BatchNorm2d
_orig_sync = nn.SyncBatchNorm if hasattr(nn, 'SyncBatchNorm') else None
class StubSyncBatchNorm(nn.BatchNorm2d):
    def __init__(self, num_features):
        super().__init__(num_features)
nn.SyncBatchNorm = StubSyncBatchNorm  # type: ignore

from timm.models.vision_transformer import Block as TimmBlock
from terratorch.models.backbones.vit_adapter_modules import (
    get_reference_points,
    deform_inputs,
    SpatialPriorModule,
    DWConv,
    ConvFFN,
    Extractor,
    Injector,
    InteractionBlock,
)

# ---- Utility fixtures ----
@pytest.fixture(scope='module')
def device():
    return torch.device('cuda' if torch.cuda.is_available() else 'cpu')

@pytest.fixture
def base_image(device):
    return torch.randn(2, 3, 64, 64, device=device)

# ---- Tests for helper functions ----

def test_get_reference_points_single_level(device):
    pts = get_reference_points([(8, 8)], device)
    assert pts.shape == (1, 64, 1, 2)
    # values normalized between 0 and 1 approximately
    assert (pts[..., 0] >= 0).all() and (pts[..., 0] <= 1).all()


def test_get_reference_points_multi_level(device):
    pts = get_reference_points([(4, 4), (2, 2)], device)
    # total points = 16 + 4 = 20
    assert pts.shape == (1, 20, 1, 2)


def test_deform_inputs_shapes(base_image):
    di1, di2 = deform_inputs(base_image)
    # Each deform_inputs list: [reference_points, spatial_shapes, level_start_index]
    assert len(di1) == 3 and len(di2) == 3
    assert di1[0].dim() == 4
    assert di2[0].dim() == 4
    assert di1[1].shape[-1] == 2
    assert di2[1].shape[-1] == 2

# ---- Tests for SpatialPriorModule ----

def test_spatial_prior_forward_no_cp(device):
    m = SpatialPriorModule(inplanes=8, embed_dim=16, in_channels=3, with_cp=False).to(device)
    x = torch.randn(2, 3, 64, 64, device=device)
    c1, c2, c3, c4 = m(x)
    assert c1.shape[1] == 16
    # c2..c4 flattened to B, L, C (due to view+transpose)
    assert c2.dim() == 3 and c3.dim() == 3 and c4.dim() == 3


def test_spatial_prior_forward_with_cp(device):
    m = SpatialPriorModule(inplanes=4, embed_dim=8, in_channels=3, with_cp=True).to(device)
    x = torch.randn(2, 3, 64, 64, device=device, requires_grad=True)
    c1, c2, c3, c4 = m(x)
    loss = c1.mean() + c2.mean() + c3.mean() + c4.mean()
    loss.backward()
    assert x.grad is not None

# ---- Tests for DWConv & ConvFFN ----

def test_dwconv_token_reconstruction(device):
    dim = 6
    N = 21
    H = W = 2
    x = torch.randn(3, N, dim, device=device)
    dw = DWConv(dim).to(device)
    out = dw(x, H, W)
    # Output tokens expected same shape tokens combining partitions
    assert out.shape[0] == 3 and out.shape[2] == dim
    assert out.shape[1] == N  # should preserve token count (16+4+1)


def test_convffn_end_to_end(device):
    ffn = ConvFFN(in_features=10, hidden_features=20, out_features=8, act_layer=nn.GELU, drop=0.1).to(device)
    x = torch.randn(2, 21, 10, device=device)
    out = ffn(x, H=2, W=2)
    assert out.shape == (2, 21, 8)

# ---- Tests for Extractor ----

def _make_deform_inputs_for_extractor(device, dim):
    # Minimal shapes for extractor: spatial_shapes one level
    spatial_shapes = torch.as_tensor([(4, 4)], device=device)
    level_start_index = torch.cat((spatial_shapes.new_zeros((1,)), spatial_shapes.prod(1).cumsum(0)[:-1]))
    ref = get_reference_points([(4, 4)], device)
    return ref, spatial_shapes, level_start_index


def test_extractor_without_cffn(device):
    dim = 12
    ext = Extractor(dim=dim, num_heads=3, n_points=4, n_levels=1, deform_ratio=0.5,
                    with_cffn=False, cffn_ratio=1.0, drop=0.0, drop_path=0.0,
                    norm_layer=nn.LayerNorm, with_cp=False).to(device)
    query = torch.randn(2, 21, dim, device=device)
    feat = torch.randn(2, 21, dim, device=device)
    ref, spatial_shapes, lsi = _make_deform_inputs_for_extractor(device, dim)
    out = ext(query, ref, feat, spatial_shapes, lsi, H=2, W=2)
    assert out.shape == query.shape


def test_extractor_with_cffn_and_cp(device):
    dim = 8
    ext = Extractor(dim=dim, num_heads=2, n_points=4, n_levels=1, deform_ratio=1.0,
                    with_cffn=True, cffn_ratio=2.0, drop=0.1, drop_path=0.2,
                    norm_layer=nn.LayerNorm, with_cp=True).to(device)
    query = torch.randn(2, 21, dim, device=device, requires_grad=True)
    feat = torch.randn(2, 21, dim, device=device)
    ref, spatial_shapes, lsi = _make_deform_inputs_for_extractor(device, dim)
    out = ext(query, ref, feat, spatial_shapes, lsi, H=2, W=2)
    loss = out.mean()
    loss.backward()
    assert query.grad is not None

# ---- Tests for Injector ----

def test_injector_forward_and_gamma_grad(device):
    dim = 10
    inj = Injector(dim=dim, num_heads=2, n_points=4, n_levels=1, deform_ratio=0.75,
                   norm_layer=nn.LayerNorm, init_values=0.5, with_cp=False).to(device)
    query = torch.randn(2, 21, dim, device=device, requires_grad=True)
    feat = torch.randn(2, 21, dim, device=device)
    spatial_shapes = torch.as_tensor([(4, 4)], device=device)
    lsi = torch.cat((spatial_shapes.new_zeros((1,)), spatial_shapes.prod(1).cumsum(0)[:-1]))
    ref = get_reference_points([(4, 4)], device)
    out = inj(query, ref, feat, spatial_shapes, lsi)
    loss = out.mean()
    loss.backward()
    assert inj.gamma.grad is not None


def test_injector_with_checkpoint(device):
    dim = 6
    inj = Injector(dim=dim, num_heads=2, n_points=4, n_levels=1, deform_ratio=0.3,
                   norm_layer=nn.LayerNorm, init_values=1.0, with_cp=True).to(device)
    query = torch.randn(2, 21, dim, device=device, requires_grad=True)
    feat = torch.randn(2, 21, dim, device=device)
    spatial_shapes = torch.as_tensor([(4, 4)], device=device)
    lsi = torch.cat((spatial_shapes.new_zeros((1,)), spatial_shapes.prod(1).cumsum(0)[:-1]))
    ref = get_reference_points([(4, 4)], device)
    out = inj(query, ref, feat, spatial_shapes, lsi)
    assert out.shape == query.shape

# ---- Tests for InteractionBlock ----

def _make_vit_blocks(dim, num=2):
    # Use timm Block expecting (B,N,C)
    return [TimmBlock(dim=dim, num_heads=2, mlp_ratio=2.0) for _ in range(num)]


def test_interaction_block_basic(device):
    dim = 8
    ib = InteractionBlock(dim=dim, num_heads=2, n_points=4, norm_layer=nn.LayerNorm,
                          drop=0.0, drop_path=0.0, with_cffn=True, cffn_ratio=2.0,
                          init_values=0.5, deform_ratio=0.7, extra_extractor=False, with_cp=False).to(device)
    x = torch.randn(2, 21, dim, device=device)
    c = torch.randn(2, 21, dim, device=device)
    deform1, deform2 = deform_inputs(torch.randn(2, dim, 64, 64, device=device))
    blocks = _make_vit_blocks(dim, num=2)
    out_x, out_c = ib(x, c, blocks, deform1, deform2, H=2, W=2)
    assert out_x.shape == x.shape and out_c.shape == c.shape


def test_interaction_block_with_extra_extractors(device):
    dim = 6
    ib = InteractionBlock(dim=dim, num_heads=2, n_points=4, norm_layer=nn.LayerNorm,
                          drop=0.1, drop_path=0.2, with_cffn=True, cffn_ratio=1.5,
                          init_values=1.0, deform_ratio=0.5, extra_extractor=True, with_cp=False).to(device)
    x = torch.randn(2, 21, dim, device=device)
    c = torch.randn(2, 21, dim, device=device)
    deform1, deform2 = deform_inputs(torch.randn(2, dim, 64, 64, device=device))
    blocks = _make_vit_blocks(dim, num=3)
    out_x, out_c = ib(x, c, blocks, deform1, deform2, H=2, W=2)
    assert out_c.shape == c.shape


def test_interaction_block_checkpoint_paths(device):
    # Use dim divisible by num_heads to satisfy timm Block assertion
    dim = 6
    ib = InteractionBlock(dim=dim, num_heads=2, n_points=2, norm_layer=nn.LayerNorm,
                          drop=0.0, drop_path=0.0, with_cffn=False, cffn_ratio=1.0,
                          init_values=0.2, deform_ratio=0.4, extra_extractor=False, with_cp=True).to(device)
    x = torch.randn(2, 21, dim, device=device, requires_grad=True)
    c = torch.randn(2, 21, dim, device=device, requires_grad=True)
    deform1, deform2 = deform_inputs(torch.randn(2, dim, 64, 64, device=device))
    blocks = _make_vit_blocks(dim, num=1)
    out_x, out_c = ib(x, c, blocks, deform1, deform2, H=2, W=2)
    (out_x.mean() + out_c.mean()).backward()
    assert x.grad is not None and c.grad is not None

# ---- Edge case tests ----

def test_dwconv_minimal_tokens_multiple_n(device):
    # Choose H,W so that partition shapes are valid for n>1.
    # For n=2: first segment tokens=32 => (H*2)*(W*2)=32 -> H*W=8 -> choose H=4,W=2
    # Second segment tokens=8 => H*W=8 OK. Third segment tokens=2 => (H//2)*(W//2)= (4//2)*(2//2)=4*1=4 mismatch with 2 so DWConv design assumes n=1 for simple scaling.
    # We'll keep n=1 for valid shapes; test already covers multi-n mismatch by expecting RuntimeError.
    dim = 4
    N = 21  # n=1
    H = W = 2
    x = torch.randn(1, N, dim)
    dw = DWConv(dim)
    out = dw(x, H, W)
    assert out.shape[1] == N


def test_convffn_defaults_hidden_out_same(device):
    ffn = ConvFFN(in_features=7, hidden_features=None, out_features=None, act_layer=nn.GELU, drop=0.0)
    x = torch.randn(1, 21, 7)
    out = ffn(x, H=2, W=2)
    assert out.shape == x.shape

if __name__ == '__main__':
    pytest.main([__file__, '-v'])



from typing import Any

import torch
from segmentation_models_pytorch.base.initialization import initialize_decoder
from segmentation_models_pytorch.decoders.unet.decoder import UnetDecoder
from torch import nn

from terratorch.registry import TERRATORCH_DECODER_REGISTRY


@TERRATORCH_DECODER_REGISTRY.register
class UNetDecoder(nn.Module):
    """UNetDecoder. Wrapper around UNetDecoder from segmentation_models_pytorch to avoid ignoring the first layer."""

    def __init__(
        self,
        embed_dim: list[int],
        channels: list[int],
        use_batchnorm: bool | str | dict[str, Any] = "batchnorm",
        attention_type: str | None = None,
    ):
        """Constructor

        Args:
            embed_dim (list[int]): Input embedding dimension for each input.
            channels (list[int]): Channels used in the decoder.
            use_batchnorm (bool, optional): Whether to use batchnorm. Defaults to True.
            attention_type (str | None, optional): Attention type to use. Defaults to None
        """
        if len(embed_dim) != len(channels):
            msg = "channels should have the same length as embed_dim"
            raise ValueError(msg)
        super().__init__()
        self.decoder = UnetDecoder(
            encoder_channels=[embed_dim[0], *embed_dim],
            decoder_channels=channels,
            n_blocks=len(channels),
            use_norm=use_batchnorm,
            add_center_block=False,
            attention_type=attention_type,
        )
        initialize_decoder(self.decoder)
        self.out_channels = channels[-1]

    def forward(self, x: list[torch.Tensor]) -> torch.Tensor:
        # The first layer is ignored in the original UnetDecoder, so we need to duplicate the first layer
        x = [x[0].clone(), *x]
        if torch.mps.is_available():
            # Fix issue on MacBooks, see https://github.com/terrastackai/terratorch/issues/859
            x = [e.contiguous() for e in x]
        return self.decoder(x)



# Copyright contributors to the Terratorch project
import os
import sys
import warnings
import pytest
import timm
import torch
import gc 
from terratorch.models.backbones import scalemae
from terratorch.registry import BACKBONE_REGISTRY
import terratorch.models.backbones.torchgeo_vit as torchgeo_vit
from terratorch.cli_tools import build_lightning_cli

NUM_CHANNELS = 6
NUM_FRAMES = 4

@pytest.fixture
def input_224():
    return torch.ones((1, NUM_CHANNELS, 224, 224))

def test_custom_module(input_224):

    sys.path.append("examples/utils/custom_modules")

    from alexnet import alexnet_encoder

    model = BACKBONE_REGISTRY.build("alexnet_encoder", num_channels=6)
    output = model(input_224)

@pytest.mark.parametrize("case", ["fit", "test", "validate"])
def test_custom_module_yaml(case):
    command_list = [case, "-c", f"examples/utils/alexnet_custom_model_config.yaml"]
    _ = build_lightning_cli(command_list)

    gc.collect()





# Copyright 2025 IBM Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import warnings
import torch
import logging
from terratorch.registry import TERRATORCH_BACKBONE_REGISTRY
from huggingface_hub import hf_hub_download

logger = logging.getLogger("terramind")

try:
    from .vqvae_backbone import VQBackbone
    vqvae_available = True
    import_error = None
except Exception as e:
    logger.debug(f"Could not import TerraMind tokenizer due to ImportError({e})")
    vqvae_available = False
    import_error = e


# Model definitions
__all__ = [
    "terramind_v1_tokenizer_s2l2a",
    "terramind_v1_tokenizer_s1rtc",
    "terramind_v1_tokenizer_s1grd",
    "terramind_v1_tokenizer_dem",
    "terramind_v1_tokenizer_lulc",
    "terramind_v1_tokenizer_ndvi",
]

pretrained_weights = {
    "terramind_v1_tokenizer_s2l2a": {
        "hf_hub_id": "ibm-esa-geospatial/TerraMind-1.0-Tokenizer-S2L2A",
        "hf_hub_filename": "TerraMind_Tokenizer_S2L2A.pt",
    },
    "terramind_v1_tokenizer_s1rtc": {
        "hf_hub_id": "ibm-esa-geospatial/TerraMind-1.0-Tokenizer-S1RTC",
        "hf_hub_filename": "TerraMind_Tokenizer_S1RTC.pt",
    },
    "terramind_v1_tokenizer_s1grd": {
        "hf_hub_id": "ibm-esa-geospatial/TerraMind-1.0-Tokenizer-S1GRD",
        "hf_hub_filename": "TerraMind_Tokenizer_S1GRD.pt",
    },
    "terramind_v1_tokenizer_dem": {
        "hf_hub_id": "ibm-esa-geospatial/TerraMind-1.0-Tokenizer-DEM",
        "hf_hub_filename": "TerraMind_Tokenizer_DEM.pt",
    },
    "terramind_v1_tokenizer_lulc": {
        "hf_hub_id": "ibm-esa-geospatial/TerraMind-1.0-Tokenizer-LULC",
        "hf_hub_filename": "TerraMind_Tokenizer_LULC.pt",
    },
    "terramind_v1_tokenizer_ndvi": {
        "hf_hub_id": "ibm-esa-geospatial/TerraMind-1.0-Tokenizer-NDVI",
        "hf_hub_filename": "TerraMind_Tokenizer_NDVI.pt",
    },
}

def checkpoint_filter_fn(state_dict) -> dict:
    """Manually filter pre-trained weights for tokenizer backbone to enable strict weight loading."""

    for k in list(state_dict.keys()):
        if k.startswith("decoder."):
            _ = state_dict.pop(k)

    return state_dict

def build_vqvae(
        variant: str = None,
        pretrained: bool = False,
        ckpt_path: str | None = None,
        **kwargs):

    if not vqvae_available:
        warnings.warn(f"Cannot import VQBackbone from terramind. "
                      f"\nMake sure to install `pip install diffusers==0.30.0`.")
        raise import_error

    model = VQBackbone(**kwargs)

    if ckpt_path is not None:
        # Load model from checkpoint
        state_dict = torch.load(ckpt_path, map_location="cpu", weights_only=True)
        state_dict = checkpoint_filter_fn(state_dict)
        loaded_keys = model.load_state_dict(state_dict, strict=False)
        if loaded_keys.missing_keys:
            logger.warning(f"Missing keys in ckpt_path {ckpt_path}: {loaded_keys.missing_keys}")
        if loaded_keys.unexpected_keys:
            logger.warning(f"Missing keys in ckpt_path {ckpt_path}: {loaded_keys.missing_keys}")

    elif pretrained:
        # Load model from Hugging Face
        state_dict_file = hf_hub_download(repo_id=pretrained_weights[variant]["hf_hub_id"],
                                          filename=pretrained_weights[variant]["hf_hub_filename"])
        state_dict = torch.load(state_dict_file, map_location="cpu", weights_only=True)
        state_dict = checkpoint_filter_fn(state_dict)
        model.load_state_dict(state_dict, strict=True)

    return model


@TERRATORCH_BACKBONE_REGISTRY.register
def terramind_v1_tokenizer_s2l2a(**kwargs):
    """
    Backbone from the S2L2A Tokenizer for TerraMind v1.
    """
    tokenizer = build_vqvae(
        variant="terramind_v1_tokenizer_s2l2a",
        image_size=256,
        n_channels=12,
        encoder_type="vit_b_enc",
        decoder_type="unet_patched",
        prediction_type="sample",
        post_mlp=True,
        patch_size=16,
        patch_size_dec=4,
        quant_type="fsq",
        codebook_size="8-8-8-6-5",
        latent_dim=5,
        clip_sample=True,
        **kwargs
    )

    return tokenizer


@TERRATORCH_BACKBONE_REGISTRY.register
def terramind_v1_tokenizer_s1rtc(**kwargs):
    """
    Backbone from the S1RTC Tokenizer for TerraMind v1.
    """
    
    tokenizer = build_vqvae(
        variant="terramind_v1_tokenizer_s1rtc",
        image_size=256,
        n_channels=2,
        encoder_type="vit_b_enc",
        decoder_type="unet_patched",
        prediction_type="sample",
        post_mlp=True,
        patch_size=16,
        patch_size_dec=4,
        quant_type="fsq",
        codebook_size="8-8-8-6-5",
        latent_dim=5,
        clip_sample=True,
        **kwargs
    )

    return tokenizer

@TERRATORCH_BACKBONE_REGISTRY.register
def terramind_v1_tokenizer_s1grd(**kwargs):
    """
    Backbone from the S1GRD Tokenizer for TerraMind v1.
    """
    
    tokenizer = build_vqvae(
        variant="terramind_v1_tokenizer_s1grd",
        image_size=256,
        n_channels=2,
        encoder_type="vit_b_enc",
        decoder_type="unet_patched",
        prediction_type="sample",
        post_mlp=True,
        patch_size=16,
        patch_size_dec=4,
        quant_type="fsq",
        codebook_size="8-8-8-6-5",
        latent_dim=5,
        clip_sample=True,
        **kwargs
    )

    return tokenizer


@TERRATORCH_BACKBONE_REGISTRY.register
def terramind_v1_tokenizer_dem(**kwargs):
    """
    Backbone from the DEM Tokenizer for TerraMind v1.
    """
    
    tokenizer = build_vqvae(
        variant="terramind_v1_tokenizer_dem",
        image_size=256,
        n_channels=1,
        encoder_type="vit_b_enc",
        decoder_type="unet_patched",
        prediction_type="sample",
        post_mlp=True,
        patch_size=16,
        patch_size_dec=4,
        quant_type="fsq",
        codebook_size="8-8-8-6-5",
        latent_dim=5,
        clip_sample=True,
        **kwargs
    )

    return tokenizer


@TERRATORCH_BACKBONE_REGISTRY.register
def terramind_v1_tokenizer_lulc(**kwargs):
    """
    Backbone from the LULC Tokenizer for TerraMind v1.
    """

    tokenizer = build_vqvae(
        variant="terramind_v1_tokenizer_lulc",
        image_size=256,
        n_channels=10,
        encoder_type="vit_b_enc",
        decoder_type="vit_b_dec",
        prediction_type="sample",
        post_mlp=True,
        patch_size=16,
        quant_type="fsq",
        codebook_size="7-5-5-5-5",
        latent_dim=5,
        **kwargs
    )

    return tokenizer



@TERRATORCH_BACKBONE_REGISTRY.register
def terramind_v1_tokenizer_ndvi(**kwargs):
    """
    Backbone from the NDVI Tokenizer for TerraMind v1.
    """

    tokenizer = build_vqvae(
        variant="terramind_v1_tokenizer_ndvi",
        image_size=256,
        n_channels=1,
        encoder_type="vit_b_enc",
        decoder_type="unet_patched",
        prediction_type="sample",
        post_mlp=True,
        patch_size=16,
        patch_size_dec=4,
        quant_type="fsq",
        codebook_size="8-8-8-6-5",
        latent_dim=5,
        clip_sample=True,
        **kwargs
    )

    return tokenizer



import tempfile
import warnings
from pathlib import Path
from unittest.mock import MagicMock, patch

import numpy as np
import pandas as pd
import pytest
import rasterio
import torch
import torch.nn as nn
import json

from terratorch.models.utils import TemporalWrapper
from terratorch.tasks.embedding_generation import EmbeddingGenerationTask

@pytest.fixture
def temp_dir():
    with tempfile.TemporaryDirectory() as tmpdir:
        yield tmpdir

@torch.no_grad()
def test_predict_step_non_temporal(tmp_path):
    task = EmbeddingGenerationTask(
        model_args={
            "backbone": "terramind_v1_tiny",
            "backbone_pretrained": False,
            "backbone_modalities": ["S2L2A"],
        },
        output_dir=str(tmp_path),
        output_format="parquet_joint",
        layers=[-1],
        embedding_pooling="mean",
    )

    B, C, H, W = 3, 12, 64, 64
    batch = {
        "image": torch.randn(B, C, H, W),
        "filename": [f"s{i}.tif" for i in range(B)],
        "metadata": {"time": ["2023-01-01"] * B},
    }

    task.predict_step(batch)
    task.on_predict_end()  # ensure parquet writer is closed/flushed

    out = tmp_path / "layer_00" / "embeddings.parquet"
    assert out.exists()

    df = pd.read_parquet(out)
    assert len(df) == B
    assert "embedding" in df.columns

    cfg = tmp_path / "configuration_summary.json"
    assert cfg.exists()

    with open(cfg, "r") as f:
        cfgj = json.load(f)
    assert cfgj["output_format"] == "parquet_joint"
    assert cfgj["n_layers_saved"] == 1


@torch.no_grad()
def test_predict_step_temporal(tmp_path):
    task = EmbeddingGenerationTask(
        model_args={"backbone": "terramind_v1_tiny",
                    "backbone_modalities": ["S2L2A"],
                    "backbone_use_temporal": True,
                    "backbone_temporal_pooling": "keep"},
        output_dir=str(tmp_path),
        output_format="parquet",
        layers=[-1],
        embedding_pooling="mean",
    )

    B, T, C, H, W = 2, 3, 12, 128, 128
    x = torch.randn(B, C, T, H, W)
    file_ids = [[f"s{b}_t{t}" for t in range(T)] for b in range(B)]
    batch = {"image": x, "filename": file_ids}
    task.predict_step(batch)

    out = tmp_path / "layer_00" / f"{file_ids[0][0]}_embedding.parquet"
    assert out.exists()
    df = pd.read_parquet(out)
    assert len(df) == 1
    assert df['embedding'][0].size == 192


@torch.no_grad()
@pytest.mark.filterwarnings("ignore::rasterio.errors.NotGeoreferencedWarning")
def test_predict_step_non_temporal_tiff_no_pooling(tmp_path):
    task = EmbeddingGenerationTask(
        model_args={
            "backbone": "terramind_v1_tiny",
            "backbone_modalities": ["S2L2A"],
        },
        output_dir=str(tmp_path),
        output_format="tiff",
        layers=[-1],
        embedding_pooling=None,   # no pooling
    )

    B, C, H, W = 2, 12, 128, 128
    x = torch.randn(B, C, H, W)
    file_ids = [f"s{b}" for b in range(B)]

    batch = {
        "image": x,
        "filename": file_ids,
    }

    task.predict_step(batch)
    out = tmp_path / "layer_00" / f"{file_ids[0]}_embedding.tif"
    assert out.exists()

    with rasterio.open(out) as ds:
        data = ds.read()

    assert data.ndim == 3
    assert data.shape[0] == 192


@pytest.fixture
def mock_backbone_registry(request):
    with patch("terratorch.models.utils.BACKBONE_REGISTRY") as registry:
        backbone = MagicMock(spec=nn.Module, name="mock_backbone_model")
        backbone.out_channels = [384]
        registry.build.return_value = backbone
        yield registry


def _has_warning(ws, substr: str) -> bool:
    return any(substr in str(w.message) for w in ws)

def test_init(temp_dir, mock_backbone_registry):
    with patch("terratorch.tasks.embedding_generation.logger") as log:
        task = EmbeddingGenerationTask(
            model_args={"backbone": "dummy"},
            output_dir=temp_dir,
            layers=[-1],
            output_format="tiff",
        )

    assert task.output_path == Path(temp_dir)
    assert task.embed_file_key == "filename"
    assert task.has_cls is False
    assert task.embedding_pooling is None
    assert task.embedding_indices == [-1]
    assert task.output_format == "tiff"

    assert task.model_args["necks"][0] == {"name": "SelectIndices", "indices": [-1]}
    assert task.model_args["necks"][1] == {"name": "ReshapeTokensToImage", "remove_cls_token": False}

    task = EmbeddingGenerationTask(
        model_args={"backbone": "dummy"},
        output_dir=temp_dir,
        layers=[0, -1],
        output_format="parquet",
        embedding_pooling="keep",
    )

    assert task.output_format == "parquet"
    assert task.model_args["necks"] == [{"name": "SelectIndices", "indices": [0, -1]}]

    with warnings.catch_warnings(record=True) as caught:
        warnings.simplefilter("always")
        task = EmbeddingGenerationTask(
            model_args={"backbone": "dummy"},
            output_dir=temp_dir,
            layers=[-1],
            output_format="tiff",
            has_cls=True,
            embedding_pooling="mean",
        )

    assert task.model_args["necks"] == [{
        "name": "AggregateTokens",
        "pooling": "mean",
        "indices": [-1],
        "drop_cls": True,
    }]
    assert any("GeoTIFF output not recommended" in str(w.message) for w in caught)
    log.info.assert_called()

def test_init_unsupported_output_format_raises(temp_dir, mock_backbone_registry):
    with pytest.raises(ValueError, match="Unsupported output format"):
        EmbeddingGenerationTask(
            model_args={"backbone": "dummy"},
            output_dir=temp_dir,
            output_format="npy",
        )

def test_init_unsupported_pooling_raises(temp_dir, mock_backbone_registry):
    with pytest.raises(ValueError, match="EmbeddingPooling .* is not supported"):
        EmbeddingGenerationTask(
            model_args={"backbone": "dummy"},
            output_dir=temp_dir,
            embedding_pooling="median",
        )

def test_output_format_is_lowercased(temp_dir, mock_backbone_registry):
    task = EmbeddingGenerationTask(
        model_args={"backbone": "dummy"},
        output_dir=temp_dir,
        output_format="PARQUET_JOINT",
    )
    assert task.output_format == "parquet_joint"

def test_infer_bt_tensor_and_dict(mock_backbone_registry):
    task = EmbeddingGenerationTask(model_args={"backbone": "dummy"})

    x = torch.randn(2, 3, 224, 224)
    assert task.infer_BT(x) == (2, 1)

    x_dict = {"optical": torch.randn(2, 3, 4, 224, 224)}
    assert task.infer_BT(x_dict) == (2, 4)


def test_check_file_ids_valid_and_errors(mock_backbone_registry):
    task = EmbeddingGenerationTask(model_args={"backbone": "dummy"})
    x_4d = torch.randn(2, 3, 224, 224)

    # valid list (B,)
    task.check_file_ids(["0", "1"], x_4d)

    # invalid shape
    with pytest.raises(ValueError, match="length mismatch"):
        task.check_file_ids(["0", "1", "2"], x_4d)

    # invalid type
    with pytest.raises(
        TypeError,
        match="must be a tensor/ndarray or a \\(nested\\) list/tuple",
    ):
        task.check_file_ids("bad", x_4d)

    # valid nested list temporal
    x_5d = torch.randn(2, 3, 3, 224, 224)
    file_ids = [["t1", "t2", "t3"], ["t4", "t5", "t6"]]
    task.check_file_ids(file_ids, x_5d)

def test_check_file_ids_invalid_inner_length(mock_backbone_registry):
    task = EmbeddingGenerationTask(model_args={"backbone": "dummy"})
    x = torch.randn(2, 3, 4, 224, 224)  # T = 4
    file_ids = [["t1", "t2"], ["t3", "t4"]]  # inner length 2 != 4

    with pytest.raises(ValueError, match="inner length 4"):
        task.check_file_ids(file_ids, x)

def test_save_configuration_summary_writes_file(tmp_path, mock_backbone_registry):
    fake_encoder = MagicMock()
    fake_encoder.return_value = [
        torch.zeros(1, 16, 8),   # layer 0
        torch.zeros(1, 32, 8),   # layer 1
    ]

    fake_model = MagicMock()
    fake_model.encoder = fake_encoder
    fake_model.parameters.return_value = [
        torch.nn.Parameter(torch.zeros(10)),
        torch.nn.Parameter(torch.zeros(20)),
    ]

    task = MagicMock()
    task._config_saved = False
    task.model = fake_model
    task.embedding_indices = [-1]
    task.output_path = tmp_path
    task.output_format = "tiff"
    task.model_args = {"backbone": "dummy_backbone"}
    task.has_cls = False
    task.embedding_pooling = None

    # bind method under test
    task.save_configuration_summary = EmbeddingGenerationTask.save_configuration_summary.__get__(task)

    # --- run ---
    x = torch.zeros(1, 3, 224, 224)
    task.save_configuration_summary(x)

    # --- assertions ---
    out_file = tmp_path / "configuration_summary.json"
    assert out_file.exists()

    cfg = json.loads(out_file.read_text())

    assert cfg["backbone"] == "dummy_backbone"
    assert cfg["model_layer_count"] == 2
    assert cfg["n_layers_saved"] == 1
    assert cfg["layers"][0]["requested_index"] == -1
    assert cfg["layers"][0]["layer_number"] == 2  # resolved -1 → last layer
    assert cfg["layers"][0]["layer_output_shape"] == [32, 8]

    assert task._config_saved is True

def test_pull_metadata_full_and_empty(mock_backbone_registry):
    task = EmbeddingGenerationTask(model_args={"backbone": "dummy"})

    data = {
        "file_id": "fid",
        "time_": "2023-01-01",
        "centre_lat": 1.0,
        "centre_lon": 2.0,
        "extra": "keep",
    }
    meta = task.pull_metadata(data)
    assert meta["file_id"] == "fid"
    assert meta["time"] == "2023-01-01"
    assert meta["center_lat"] == 1.0
    assert meta["center_lon"] == 2.0
    assert "extra" in data
    assert "file_id" not in data

    data2 = {"x": 1}
    assert task.pull_metadata(data2) == {}

def test_write_parquet(temp_dir, mock_backbone_registry):
    task = EmbeddingGenerationTask(
        model_args={"backbone": "dummy"},
        output_dir=temp_dir,
        output_format="parquet",
    )
    emb = np.random.random((768, 1024))
    meta = {"time": np.array("2023-01-01")}
    task.write_parquet(emb, "p", meta, Path(temp_dir))

    out = Path(temp_dir) / "p_embedding.parquet"
    df = pd.read_parquet(out)
    assert len(df) == 1
    assert len(df["embedding"][0]) == 768
    assert len(df["embedding"][0][0]) == 1024
    assert df["time"][0] == "2023-01-01"


@pytest.mark.filterwarnings("ignore::rasterio.errors.NotGeoreferencedWarning")
def test_write_tiff(temp_dir, mock_backbone_registry):
    task = EmbeddingGenerationTask(
        model_args={"backbone": "dummy"},
        output_dir=temp_dir,
        output_format="tiff",
    )

    emb = np.random.random((4, 8, 8)).astype(np.float32)
    meta = {"time": "2023-01-01"}

    task.write_tiff(emb, "p.tif", meta, Path(temp_dir))

    out = Path(temp_dir) / "p_embedding.tif"
    assert out.exists()

    with rasterio.open(out) as ds:
        data = ds.read()
        tags = ds.tags()

    assert data.shape == (4, 8, 8)
    assert tags["time"] == "2023-01-01"

def test_write_parquet_batch_and_close_parquet(temp_dir, mock_backbone_registry):
    task = EmbeddingGenerationTask(
        model_args={"backbone": "dummy"},
        output_dir=temp_dir,
        output_format="parquet_joint",
    )

    task.write_parquet_batch(
        emb_np=np.ones((4,1024)),
        file_ids=["a", "b", "c", "d"],
        metadata={},
        is_temporal=False,
        dir_path=Path(temp_dir) / "layer_00",
    )
    task.join_parquet_files()

    out = Path(temp_dir) / "layer_00" / "embeddings.parquet"
    assert out.exists()

    df = pd.read_parquet(out)
    assert len(df) == 4
    assert df["file_id"][0] == "a"
    assert len(df["embedding"][0]) == 1024

def test_predict_step_missing_filename(temp_dir, mock_backbone_registry):
    model = MagicMock()
    model.return_value = torch.randn(2, 4, 3, 3)

    with patch("terratorch.models.utils.BACKBONE_REGISTRY") as reg:
        reg.build.return_value = model
        task = EmbeddingGenerationTask(model_args={"backbone": "dummy"}, output_dir=temp_dir)

    batch = {
        "image": torch.randn(2, 3, 224, 224),
        "time": ["2023-01-01", "2023-01-02"],
    }
    with pytest.raises(KeyError, match="not found in input dictionary"):
        task.predict_step(batch)

if __name__ == '__main__':
    pytest.main([__file__, '-v', '--tb=short'])



from terratorch.registry.registry import (  # noqa: I001 dont sort, import order matters
    MultiSourceRegistry,
    Registry,
    BACKBONE_REGISTRY,
    DECODER_REGISTRY,
    NECK_REGISTRY,
    FULL_MODEL_REGISTRY,
    TERRATORCH_BACKBONE_REGISTRY,
    TERRATORCH_DECODER_REGISTRY,
    TERRATORCH_NECK_REGISTRY,
    TERRATORCH_FULL_MODEL_REGISTRY,
    MODEL_FACTORY_REGISTRY,
    explore_model_card,
)
import terratorch.registry.smp_registry  # register smp registry
import terratorch.registry.timm_registry  # register timm registry
import terratorch.registry.mmseg_registry
import terratorch.registry.custom_registry

__all__ = [
    "BACKBONE_REGISTRY",
    "DECODER_REGISTRY",
    "FULL_MODEL_REGISTRY",
    "MODEL_FACTORY_REGISTRY",
    "NECK_REGISTRY",
    "TERRATORCH_BACKBONE_REGISTRY",
    "TERRATORCH_DECODER_REGISTRY",
    "TERRATORCH_FULL_MODEL_REGISTRY",
    "TERRATORCH_NECK_REGISTRY",
    "MultiSourceRegistry",
    "Registry",
]



# Copyright 2025 IBM Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# ---
#
# This project includes code adapted from the original work by EPFL and Apple Inc.,
# licensed under the Apache License, Version 2.0.
# Source: https://github.com/apple/ml-4m/

import copy
import math
import warnings

import numpy as np
import torch
import torch.nn.functional as F
from einops import rearrange, repeat
from tqdm import tqdm
from torch import nn
from collections import defaultdict


def get_sentinel_to_id_mapping(tokenizer, match_str="[S_"):
    sentinel_tokens = {k: v for k, v in tokenizer.get_vocab().items() if k.startswith(match_str)}
    # Extract the sentinel token id, the id is of the form "[S_0]", "[S_1]", etc.
    sentinel_to_id = {int(k.split("_")[1][:-1]): v for k, v in sorted(sentinel_tokens.items(), key=lambda x:x[1])}
    return sentinel_to_id


def split_by_sentinel(seq_ids, sentinel_ids):
    splits = defaultdict(list)
    cur_sentinel = None
    for token in seq_ids:
        if token in sentinel_ids:
            cur_sentinel = token
        else:
            splits[cur_sentinel].append(token)

    return splits


def merge_span_masking(input_seq, decoder_seq, sentinel_ids):
    decoder_splits = split_by_sentinel(decoder_seq, sentinel_ids)
    out_seq = []
    for token in input_seq:
        if token in sentinel_ids:
            out_seq.extend(decoder_splits[token])
        else:
            out_seq.append(token)
    return out_seq


def cosine_schedule(num_steps, total_tokens):
    iters = np.arange(num_steps)
    base_value = 1
    final_value = 0
    schedule = np.array(
        [final_value + 0.5 * (base_value - final_value) * (1 + math.cos(math.pi * i / (len(iters)))) for i in iters])
    schedule_tokens = [round(total_tokens * i) for i in (schedule[:-1] - schedule[1:])]
    schedule_tokens.append(total_tokens - sum(schedule_tokens))
    return np.array(schedule_tokens)


def linear_schedule(num_steps, total_tokens):
    schedule = np.linspace(0, total_tokens, num_steps + 1, dtype=int)
    schedule_tokens = np.diff(schedule)[::-1]
    schedule_tokens.sort()  # Sorts the array in ascending order.
    schedule_tokens = schedule_tokens[::-1]  # Reverses the array to descending order.
    return np.trim_zeros(schedule_tokens, 'b')  # Trims trailing zeros.


def onex_temp_schedule(max_t, min_t, token_schedule, power=0.5, min_linspace=1, max_linspace=100):
    """Abitrary temperature schedule for one over x"""
    x = np.linspace(min_linspace, max_linspace, num=sum(token_schedule))
    y = 1 / (x ** power)
    y = y - min(y)
    y = y / max(y)
    unscaled_schedule = y
    schedule_cumsum = np.cumsum(token_schedule) / np.sum(token_schedule)
    unscaled_schedule = [(1 - cs) * us for us, cs in zip(unscaled_schedule, schedule_cumsum)]

    temp_schedule = np.array([min_t + (max_t - min_t) * s for s in unscaled_schedule]).clip(min=1e-9)
    return temp_schedule


def linear_temp_schedule(temp, token_schedule):
    """ Temperature that decays the temperature inversely proportional to the token schedule. """
    return np.concatenate([np.array([temp * 1.0]),
                           (temp * (token_schedule.sum() - token_schedule.cumsum()) / token_schedule.sum())[:-1]]).clip(
        min=1e-9)


def empty_img_modality(mod_dict, key):
    # Input mask
    mod_dict[key]['input_mask'][:] = True
    
    # Target Mask
    mod_dict[key]['target_mask'][:] = False
    
    return mod_dict

def empty_seq_modality(mod_dict, s1_id=5):
    # To create an empty sequence, we suppose an input budget of 1, and the rest assigned to targets

    # Input tensor
    # Input is [S_1], target is [S_1] ...... [S_2]
    # (so [S_1] [S_1] ..... [S_2] when combined)
    mod_dict['tensor'][:] = 0
    mod_dict['tensor'][:,[0,1]] = s1_id  # s1_id is id of the first sentinel token ([S_1])
    mod_dict['tensor'][:,-1] = s1_id + 1

    # Input mask
    # Set first token to input (i.e. 0), rest to target (i.e. 1)
    mod_dict['input_mask'][:] = True
    mod_dict['input_mask'][:,0] = False
    
    # Target Mask
    mod_dict['target_mask'] = ~mod_dict['input_mask']

    # Decoder attn mask
    # WARNING: Not needed / used in GenerationSampler, where causal mask is enforced
    # First token is input, not part of target
    mod_dict['decoder_attention_mask'][:] = 1
    mod_dict['decoder_attention_mask'][:, 0] = 0

    return mod_dict

def empty_seq_emb_modality(mod_dict):
    # It is crucial to specify the input mask as such, CFG won't work otherwise!
    mod_dict['input_mask'][:, 0] = False
    return mod_dict
    

def init_empty_target_modality(modality_info, domain, batch_size, num_tokens, device, s1_id=5):
    """
    Initializes an empty target modality dictionary for a given domain. 
    Used to initialize target modality dictionaries for generation. 
    """
    if modality_info[domain]['type'] == 'img':
        # Initialize mod dict
        mod_dict = {
            'tensor': torch.zeros((batch_size, num_tokens), dtype=torch.int64, device=device),
            'input_mask': torch.ones((batch_size, num_tokens), dtype=torch.bool, device=device),
            'target_mask': torch.zeros((batch_size, num_tokens), dtype=torch.bool, device=device),
        }

    elif modality_info[domain]['type'] in ['seq', 'seq_token', 'seq_emb']:
        # Initialize mod dict
        mod_dict = {
            'tensor': torch.zeros((batch_size, num_tokens), dtype=torch.int, device=device),
            'input_mask': torch.ones((batch_size, num_tokens), dtype=torch.bool, device=device),
            'target_mask': torch.zeros((batch_size, num_tokens), dtype=torch.bool, device=device),
            'decoder_attention_mask': torch.zeros((batch_size, num_tokens), dtype=torch.bool, device=device),
        }
        # Set it to the correct values
        if modality_info[domain]['type'] in ['seq', 'seq_token']:
            mod_dict = empty_seq_modality(mod_dict, s1_id=s1_id)
        elif modality_info[domain]['type'] == 'seq_emb':
            mod_dict = empty_seq_emb_modality(mod_dict)
    else:
        raise ValueError()
        
    return mod_dict

def init_conditioned_target_modality(mod_dict, modality_info, domain, num_target_tokens, eos_id=3, s1_id=5):
    batch_size, input_length = mod_dict["tensor"].shape[:2]
    device = mod_dict["tensor"].device

    if modality_info[domain]['type'] in ['seq', 'seq_token']:
        # Extend the input tokens with target tokens
        mod_dict["tensor"] = torch.cat([
            mod_dict["tensor"],
            torch.zeros((batch_size, num_target_tokens), dtype=torch.int, device=device)
        ], dim=1)
        mod_dict["input_mask"] = torch.cat([
            mod_dict["input_mask"],
            torch.ones((batch_size, num_target_tokens), dtype=torch.bool, device=device)
        ], dim=1)
        mod_dict["target_mask"] = torch.cat([
            mod_dict["target_mask"],
            torch.zeros((batch_size, num_target_tokens), dtype=torch.bool, device=device)
        ], dim=1)
        mod_dict["decoder_attention_mask"] = torch.cat([
            mod_dict["decoder_attention_mask"],
            torch.ones((batch_size, num_target_tokens), dtype=torch.bool, device=device)
        ], dim=1)

        if eos_id in mod_dict['tensor']:
            eos_indices = torch.where(mod_dict['tensor'] == eos_id)[1]
        else:
            warnings.warn(f'Cannot find EOS token in {domain} input, assuming last input position.')
            eos_indices = [input_length] * len(mod_dict['tensor'])
        for i in range(len(mod_dict['tensor'])):
            eos_idx = eos_indices[i]
            # Update input tensor
            # Input is [S_1], target is [S_1] ...... [S_2] (so Input [S_1] [S_1] ..... [S_2] when combined)
            mod_dict['tensor'][i, [eos_idx, eos_idx+1]] = s1_id
            mod_dict['tensor'][i, -1] = s1_id + 1

            mod_dict['input_mask'][i, :eos_idx+1] = False
            mod_dict['input_mask'][i, eos_idx+1:] = True
        mod_dict['target_mask'] = ~mod_dict['input_mask']
        mod_dict["decoder_attention_mask"] = mod_dict['input_mask']

    elif modality_info[domain]['type'] in ['img']:
        # TODO: Implement image masking in input
        raise NotImplementedError(f'Conditioned target modality not implemented for {domain}')
    else:
        raise NotImplementedError(f'Conditioned target modality not implemented for {domain}')

    return mod_dict


def init_full_input_modality(value, modality_info, domain, device, eos_id=3):
    mod_dict = value if isinstance(value, dict) else {"tensor": value}

    if 'input_mask' not in mod_dict:
        mod_dict['input_mask'] = torch.zeros(mod_dict['tensor'].shape, dtype=torch.bool, device=device)
    if 'target_mask' not in mod_dict:
        mod_dict['target_mask'] = torch.ones(mod_dict['tensor'].shape, dtype=torch.bool, device=device)
    if 'decoder_attention_mask' not in mod_dict:
        mod_dict['decoder_attention_mask'] = torch.zeros(mod_dict['tensor'].shape, dtype=torch.bool, device=device)

    if modality_info[domain]['type'] == 'img':
        mod_dict['input_mask'][:] = False
        mod_dict['target_mask'][:] = True

    elif modality_info[domain]['type'] in ['seq', 'seq_token']:
        if eos_id in mod_dict['tensor']:
            eos_idx = torch.where(mod_dict['tensor'] == eos_id)[1][0].item()
        else:
            warnings.warn(f'Cannot find EOS token in {domain} input, ignoring input.')
            mod_dict['tensor'][:,0] = eos_id
            eos_idx = 0
        mod_dict['input_mask'][:,:eos_idx+1] = False
        mod_dict['input_mask'][:,eos_idx+1:] = True
        mod_dict['target_mask'][:] = True

    elif modality_info[domain]['type'] in ['seq_emb']:
        # T5 caption has the valid mask saved alongside the embeddings
        mod_dict['input_mask'] = ~mod_dict['mask_valid']
        mod_dict['target_mask'] = torch.ones_like(mod_dict['mask_valid'])
        mod_dict['decoder_attention_mask'] = torch.zeros_like(mod_dict['mask_valid'])
        
    return mod_dict

def custom_text(input_text, device, text_tokenizer, target_max_len=50, start_token="[S_1]", eos_token="[EOS]"):
    input_ids = text_tokenizer.encode(input_text).ids
    input_ids = torch.tensor(input_ids).unsqueeze(0)

    target_text = [start_token]
    target_text.extend(["[PAD]"] * (target_max_len - 2))
    target_text.append(eos_token)
    target_text = " ".join(target_text)
    target_ids = text_tokenizer.encode(target_text).ids
    target_ids = torch.tensor(target_ids).unsqueeze(0)

    all_ids = torch.cat([input_ids, target_ids], dim=1)

    input_mask = torch.cat([
        torch.zeros_like(input_ids, dtype=torch.bool), 
        torch.ones_like(target_ids, dtype=torch.bool),
        ], dim=1)

    target_mask = torch.cat([
        torch.ones_like(input_ids, dtype=torch.bool), 
        torch.zeros_like(target_ids, dtype=torch.bool),
        ], dim=1)

    mod_dict = {}
    mod_dict['tensor'] = all_ids.to(device)
    mod_dict['input_mask'] = input_mask.to(device)
    mod_dict['target_mask'] = target_mask.to(device)
    mod_dict['decoder_attention_mask'] = torch.zeros(all_ids.shape, dtype=torch.bool, device=device)

    return mod_dict


def build_chained_generation_schedules(
        cond_domains: list[str], 
        target_domains: list[str],
        tokens_per_target: list[int],
        autoregression_schemes: list[str], 
        decoding_steps: list[int], 
        token_decoding_schedules: list[str],
        temps: list[float],
        temp_schedules: list[str],
        cfg_scales: list[float], 
        cfg_schedules: list[str],
        cfg_grow_conditioning: bool = False, 
        modality_info: dict | None = None,
    ):
    """
    Builds a list of chained generation schedules, where each schedule is a tuple of the form:
    (target_modality, schema, number of decoded tokens, temperature, guidance_scale, cfg_cond_domains)

    Args:
        cond_domains: list of conditioning domains
        target_domains: list of target domains
        tokens_per_target: list of number of tokens to decode for each target domain
        autoregression_schemes: list of autoregression schemes for each target domain. maskgit, roar, or autoregressive
        decoding_steps: list of number of maskgit steps for each target domain (if applicable)
        token_decoding_schedules: list of maskgit token schedules for each target domain (if applicable). cosine or linear
        temps: list of starting temperatures for each target domain
        temp_schedules: list of temperature schedules for each target domain. linear, constant, or onex:{min_t}:{power}
        cfg_scales: list of classifier-free guidance scales for each target domain
        cfg_schedules: list of classifier-free guidance schedules for each target domain. constant or cosine
        cfg_grow_conditioning: After every completed modality, add them to classifier-free guidance conditioning
        modality_info: Dictionary with metadata for each modality, optionally used to verify that the schedule is compatible with the modality
    """
    
    # list of {target_modality, schema, number of decoded tokens, temperature, guidance_scale, cfg_cond_domains} dicts
    chained_schedules = []
    
    cond_domains = cond_domains.copy()
    
    for target_idx in range(len(target_domains)):
        
        scheme = autoregression_schemes[target_idx]
        target_domain = target_domains[target_idx]
        ntoks = tokens_per_target[target_idx]
        maskgit_token_schedule_name = token_decoding_schedules[target_idx]
        temp = temps[target_idx]
        temp_schedule_name = temp_schedules[target_idx]
        cfg_scale = cfg_scales[target_idx]
        cfg_schedule_name = cfg_schedules[target_idx]

        # Auto-regressive (caption, detection, ...)
        if scheme == 'autoregressive':
            chained_schedules.append({
                'target_domain': target_domain,
                'scheme': scheme,
                'num_tokens': None,
                'temperature': temp,
                'cfg_scale': cfg_scale,
                'cfg_cond_domains': cond_domains.copy()
            })
            continue

        # Use modality info for (optional) assert if provided
        if modality_info is not None:
            assert modality_info[target_domain]['type'] not in ['seq', 'seq_token'], f'Illegal autoregressive scheme {scheme} for target domain {target_domain}'

        # Token schedule
        if scheme == 'maskgit':
            # MaskGIT token schedule setup
            num_steps = decoding_steps[target_idx]
            if maskgit_token_schedule_name == 'cosine':
                token_schedule = cosine_schedule(num_steps, (ntoks))
            elif maskgit_token_schedule_name == 'linear':
                token_schedule = linear_schedule(num_steps, (ntoks))
            else:
                raise ValueError(f'Illegal MaskGIT token schedule {maskgit_token_schedule_name}')
        elif scheme == 'roar':
            # ROAR token schedule setup (one-by-one, but random order)
            num_steps = decoding_steps[target_idx]            
            token_schedule = linear_schedule(num_steps, ntoks)
        else:
            raise ValueError(f'Illegal decoding scheme {scheme}')
        
        # Temperature schedule
        if temp_schedule_name == 'linear':
            temp_schedule = linear_temp_schedule(temp, token_schedule)
        elif temp_schedule_name == 'constant':
            temp_schedule = temp * np.ones(num_steps)
        elif 'onex' in temp_schedule_name:
            # onex temperature schedule has to be formatted like onex:{min_t}:{power}
            min_t, power = [float(f) for f in temp_schedule_name.split(':')[1:]]
            temp_schedule = onex_temp_schedule(max_t=temp, min_t=min_t, token_schedule=token_schedule, power=power)
        else:
            raise ValueError(f'Illegal temperature schedule {temp_schedule_name}')

        # Classifier-free guidance scale schedule
        if cfg_schedule_name == 'constant':
            if isinstance(cfg_scale, float):
                cfg_schedule = cfg_scale * np.ones(num_steps)
            elif isinstance(cfg_scale, list):
                cfg_schedule = np.array(cfg_scale) * np.ones(num_steps).reshape(-1, 1)
        elif cfg_schedule_name == 'cosine':
            raise NotImplementedError()
        else:
            raise ValueError(f'Illegal guidance schedule {cfg_schedule_name}')

        # Concatenate schedule for this modality with previous ones
        schedule = [
            {
                'target_domain': target_domain,
                'scheme': scheme,
                'num_tokens': tok,
                'temperature': temp,
                'cfg_scale': cfg,
                'cfg_cond_domains': cond_domains.copy()
            }
            for tok, temp, cfg in zip(token_schedule, temp_schedule, cfg_schedule)
        ]
        chained_schedules.extend(schedule)
        
        # Optionally add this new modality to the ones affected by classifier-free guidance
        if cfg_grow_conditioning:
            cond_domains.append(target_domain)
            
    return chained_schedules


class GenerationSampler(nn.Module):
    """Sampler that wraps a trained 4M model for generation use cases.
    Implements standard autoregressive, MaskGIT, and ROAR generation schemes with chaining and weighted guidance."""

    def __init__(self, model):
        super().__init__()
        self.model = model


    def top_k_top_p_filtering(self, logits, top_k=0.0, top_p=0.0):
        # Compatible with batching
        # From https://gist.github.com/thomwolf/1a5a29f6962089e871b94cbd09daf317
        if top_k > 0.0:
            if isinstance(top_k, int):
                k = min(top_k, logits.shape[-1])
            elif isinstance(top_k, float):
                k = min(int(top_k * logits.shape[-1]), logits.shape[-1])
            else:
                raise ValueError(f"Invalid value for top_k: {top_k}")

            # Remove all tokens with a probability less than the last token of the top-k
            indices_to_remove = logits < torch.topk(logits, k)[0][..., -1, None]
            logits[indices_to_remove] = float("-inf")

        if top_p > 0.0:
            sorted_logits, sorted_indices = torch.sort(logits, dim=1, descending=True)
            cum_probs = torch.cumsum(F.softmax(sorted_logits, dim=-1), dim=-1)
            sorted_indices_to_remove = cum_probs > top_p
            # Shift the indices to the right to keep also the first token above the threshold
            sorted_indices_to_remove[..., 1:] = sorted_indices_to_remove[..., :-1].clone()
            sorted_indices_to_remove[..., 0] = 0

            restore_indices = torch.argsort(sorted_indices, dim=-1)
            indices_to_remove = torch.gather(sorted_indices_to_remove, dim=-1, index=restore_indices)
            logits[indices_to_remove] = float("-inf")

        return logits

    def sample_tokens(self, logits, temperature=1.0, top_k=0.0, top_p=0.0):
        if np.isclose(temperature, 0, atol=1e-10):
            samples = torch.argmax(logits, dim=-1)
            # Since argmax is used, all sampled_probs will be 1 as we're selecting the max probability
            sampled_probs = torch.ones_like(samples, dtype=torch.get_default_dtype())
        else:
            filtered_logits = self.top_k_top_p_filtering(logits, top_k, top_p)
            probs = F.softmax(filtered_logits / temperature, dim=-1)
            samples = torch.multinomial(probs, 1)[:, 0]
            sampled_probs = probs[torch.arange(len(samples)), samples]
        return samples, sampled_probs
    
    def sample_tokens_batched(self, logits, temperature=1.0, top_k=0.0, top_p=0.0):
        if logits.ndim > 2:
            B, N = logits.shape[0], logits.shape[1]
            logits = rearrange(logits, 'b n v -> (b n) v')
            samples, sampled_probs = self.sample_tokens(logits, temperature, top_k, top_p)
            samples = rearrange(samples, '(b n) -> b n', b=B, n=N)
            sampled_probs = rearrange(sampled_probs, '(b n) -> b n', b=B, n=N)
            return samples, sampled_probs
        else:
            return self.sample_tokens(logits, temperature, top_k, top_p)

    def select_tokens(self, logits, num_select, temperature=1.0, top_k=0.0, top_p=0.0, return_all_samples=False):
        samples, sampled_probs = self.sample_tokens(logits, temperature, top_k, top_p)
        top_indices = torch.topk(sampled_probs, num_select)[1]
        top_samples = samples[top_indices]
        if return_all_samples:
            return top_samples, top_indices, samples
        else:
            return top_samples, top_indices
        
    def select_tokens_batched(self, logits, num_select, temperature=1.0, top_k=0.0, top_p=0.0, return_all_samples=False):
            if logits.ndim > 2:
                samples, sampled_probs = self.sample_tokens_batched(logits, temperature, top_k, top_p) # both of shape (B, N)
                top_indices = torch.topk(sampled_probs, num_select, dim=-1)[1]
                # Need to switch to gather instead of indexing here
                top_samples = torch.gather(samples, dim=-1, index=top_indices)
                if return_all_samples:
                    return top_samples, top_indices, samples
                else:
                    return top_samples, top_indices
            else:
                return self.sample_tokens(logits, num_select, temperature, top_k, top_p, return_all_samples)


    def forward_mask_encoder_generation(self, encoder_mod_dict):
        """Modification of forward_mask_encoder adapted for generation, with support for batching
        """
        # Form input
        B = list(encoder_mod_dict.values())[0]['tensor'].shape[0]

        encoder_tokens_all, emb_all, encoder_mask_all, mod_mask_all = self.model.cat_encoder_tensors(encoder_mod_dict)
        # Take max num encoder of tokens (although assuming it's the same everywhere would be better)
        num_encoder_tokens = (~encoder_mask_all.reshape(B, -1)).sum(dim=1).max()

        # Add arange multiplied by small constant to mask so they get sorted in a deterministic way
        mask_arange = torch.arange(encoder_mask_all.shape[1], device=encoder_mask_all.device).unsqueeze(0) * 1e-6
        ids_shuffle = torch.argsort(encoder_mask_all + mask_arange, dim=1)
        # ids_restore = torch.argsort(ids_shuffle, dim=1)
        ids_keep = ids_shuffle[:, :num_encoder_tokens]

        encoder_tokens = torch.gather(encoder_tokens_all, dim=1,
                                      index=repeat(ids_keep, "b n -> b n d", d=encoder_tokens_all.shape[2]))
        encoder_emb = torch.gather(emb_all, dim=1, index=repeat(ids_keep, "b n -> b n d", d=emb_all.shape[2]))
        encoder_mask = torch.gather(encoder_mask_all, dim=1, index=ids_keep)
        mod_mask = torch.gather(mod_mask_all, dim=1, index=ids_keep)

        if self.model.num_register_tokens > 0:
            register_tokens = repeat(self.model.register_tokens, '() n d -> b n d', b=B)
            # We add prompt tokens at the beginning of the sequence
            encoder_tokens = torch.cat([register_tokens, encoder_tokens], dim=1)
            encoder_emb = torch.cat([torch.zeros_like(register_tokens), encoder_emb], dim=1)
            encoder_mask = torch.cat([torch.zeros((B, register_tokens.shape[1]), dtype=torch.bool, device=encoder_mask.device), encoder_mask], dim=1)
            mod_mask = torch.cat([torch.full((B, register_tokens.shape[1]), -1, dtype=torch.int, device=mod_mask.device), mod_mask], dim=1)

        encoder_tokens[encoder_mask] = 0.
        encoder_emb[encoder_mask] = 0.
        mod_mask[encoder_mask] = -1
        # Mask could be of shape 'b n1 n2' but not needed for masked_fill
        # This means this mask can then be re-used for decoder cross-attention
        encoder_mask = rearrange(encoder_mask, 'b n2 -> b 1 n2')

        return encoder_tokens, encoder_emb, encoder_mask, mod_mask


    def forward_mask_decoder_maskgit(self, mod_dict, target_mod, seed=None):
        """Modification of forward_mask_decoder for MaskGIT generation, with support for batching
        """
        if seed is not None:
            torch.manual_seed(seed)
        d = mod_dict[target_mod]
        decoder_tokens_all = torch.zeros_like(d['x']) + self.model.mask_token
        emb_all = d['emb']
        decoder_mask_all = d['target_mask']
        B = decoder_tokens_all.shape[0] # Get batch size
        mod_mask_all = torch.full_like(d['ids'], self.model.modality_info[target_mod]['id'], dtype=torch.int)
        mod_pos_all = torch.arange(d['x'].shape[1], device=d['x'].device).unsqueeze(0)
        mod_pos_all = repeat(mod_pos_all, '1 n -> b n', b=B) # Added: Expansion for batching
        num_decoder_tokens = (~decoder_mask_all[0]).sum()  # Adapted for batching / Assumes num_decoder_tokens is the same across the batch


        # Add arange multiplied by small constant to mask so they get sorted in a deterministic way
        mask_arange = torch.arange(decoder_mask_all.shape[1], device=decoder_mask_all.device).unsqueeze(0) * 1e-6
        ids_shuffle = torch.argsort(decoder_mask_all + mask_arange, dim=1)
        # ids_restore = torch.argsort(ids_shuffle, dim=1)
        ids_keep = ids_shuffle[:, :num_decoder_tokens]

        decoder_tokens = torch.gather(decoder_tokens_all, dim=1, index=repeat(ids_keep, "b n -> b n d", d=decoder_tokens_all.shape[2]))
        decoder_emb = torch.gather(emb_all, dim=1, index=repeat(ids_keep, "b n -> b n d", d=emb_all.shape[2]))
        decoder_mask = torch.gather(decoder_mask_all, dim=1, index=ids_keep)
        mod_mask = torch.gather(mod_mask_all, dim=1, index=ids_keep)
        mod_pos = torch.gather(mod_pos_all, dim=1, index=ids_keep)

        decoder_tokens[decoder_mask] = 0.
        decoder_emb[decoder_mask] = 0.
        mod_mask[decoder_mask] = -1

        return decoder_tokens, decoder_emb, decoder_mask, mod_mask, mod_pos

    def forward_mask_decoder_roar(self, mod_dict, target_mod, num_select, seed=None):
        """Modification of forward_mask_decoder for ROAR generation, with support for batching
        """
        if seed is not None:
            torch.manual_seed(seed)
        d = mod_dict[target_mod]
        decoder_tokens_all = torch.zeros_like(d['x']) + self.model.mask_token
        emb_all = d['emb']
        decoder_mask_all = d['target_mask']
        B = decoder_tokens_all.shape[0] # Get batch size
        mod_mask_all = torch.full_like(d['ids'], self.model.modality_info[target_mod]['id'], dtype=torch.int)
        mod_pos_all = torch.arange(d['x'].shape[1], device=d['x'].device).unsqueeze(0)
        mod_pos_all = repeat(mod_pos_all, '1 n -> b n', b=B) # Added: Expansion for batching
        # Only keep the first num_select tokens
        num_decoder_tokens = min(num_select, (~decoder_mask_all[0]).sum())  # Adapted for batching / Assumes num_decoder_tokens is the same across the batch
 
        # Add a small random number to the mask so they get sorted in a random way, but keeping the masked tokens first
        mask_rand = torch.rand(decoder_mask_all.shape[1], device=decoder_mask_all.device).unsqueeze(0) * 1e-6
        ids_shuffle = torch.argsort(decoder_mask_all + mask_rand, dim=1)
        # ids_restore = torch.argsort(ids_shuffle, dim=1)
        # Only keep the first num_select_tokens
        ids_keep = ids_shuffle[:, :num_decoder_tokens]

        decoder_tokens = torch.gather(decoder_tokens_all, dim=1, index=repeat(ids_keep, "b n -> b n d", d=decoder_tokens_all.shape[2]))
        decoder_emb = torch.gather(emb_all, dim=1, index=repeat(ids_keep, "b n -> b n d", d=emb_all.shape[2]))
        decoder_mask = torch.gather(decoder_mask_all, dim=1, index=ids_keep)
        mod_mask = torch.gather(mod_mask_all, dim=1, index=ids_keep)
        mod_pos = torch.gather(mod_pos_all, dim=1, index=ids_keep)

        decoder_tokens[decoder_mask] = 0.
        decoder_emb[decoder_mask] = 0.
        mod_mask[decoder_mask] = -1

        return decoder_tokens, decoder_emb, decoder_mask, mod_mask, mod_pos

    def forward_mask_decoder_autoregressive(self, mod_dict, target_mod, seed=None):
        # Adapted for batching
        if seed is not None:
            torch.manual_seed(seed)
        # This is the concatenation part
        d = mod_dict[target_mod]
        decoder_ids_all = d['ids']
        emb_all = d['emb']
        decoder_mask_all = d['target_mask']
        B = decoder_ids_all.shape[0] # Get batch size
        mod_mask_all = torch.full_like(d['ids'], self.model.modality_info[target_mod]['id'], dtype=torch.int)
        mod_pos_all = torch.arange(d['x'].shape[1], device=d['x'].device).unsqueeze(0)
        mod_pos_all = repeat(mod_pos_all, '1 n -> b n', b=B) 
        num_decoder_tokens = (~decoder_mask_all[0]).sum() # Adapted for batching, but assumes num_decoder_tokens is the same across the batch
 
        # Add arange multiplied by small constant to mask so they get sorted in a deterministic way
        mask_arange = torch.arange(decoder_mask_all.shape[1], device=decoder_mask_all.device).unsqueeze(0) * 1e-6
        ids_shuffle = torch.argsort(decoder_mask_all + mask_arange, dim=1)
        # ids_restore = torch.argsort(ids_shuffle, dim=1)
        ids_keep = ids_shuffle[:, :num_decoder_tokens]

        # Same as in forward_mask_decoder
        decoder_ids = torch.gather(decoder_ids_all, dim=1, index=ids_keep)
        decoder_emb = torch.gather(emb_all, dim=1, index=repeat(ids_keep, "b n -> b n d", d=emb_all.shape[2]))
        decoder_mask = torch.gather(decoder_mask_all, dim=1, index=ids_keep)
        mod_mask = torch.gather(mod_mask_all, dim=1, index=ids_keep)
        mod_pos = torch.gather(mod_pos_all, dim=1, index=ids_keep)

        decoder_ids[decoder_mask] = 0
        decoder_emb[decoder_mask] = 0.
        mod_mask[decoder_mask] = -1

        return decoder_ids, decoder_emb, decoder_mask, mod_mask, mod_pos

    def merge_sequences(self, mod_dict, pred_ids, target_mod, text_tokenizer, default_sentinel="[S_1]"):
        device = mod_dict[target_mod]['tensor'].device
        # Get input ids
        input_ids = mod_dict[target_mod]['tensor'].squeeze().detach().cpu()
        input_ids = input_ids[mod_dict[target_mod]['input_mask'].squeeze().detach().cpu() == 0]
        input_ids = input_ids.tolist()

        if len(input_ids) == 0:
            input_ids = [text_tokenizer.get_vocab()[default_sentinel]]

        # Get predicted ids
        pred_ids = pred_ids.squeeze().detach().cpu().tolist()
        if isinstance(pred_ids, int):
            pred_ids = [pred_ids]

        # Get sentinel ids using the tokenizer
        sentinel_ids = set(get_sentinel_to_id_mapping(text_tokenizer).values())
        # Perform merging
        merged_ids = merge_span_masking(input_ids, pred_ids, sentinel_ids)
        merged_ids = torch.tensor(merged_ids).unsqueeze(0)
        # Create new dict
        new_input_mask = torch.zeros_like(merged_ids, dtype=torch.bool)
        new_target_mask = torch.ones_like(merged_ids, dtype=torch.bool)
        new_dict = {'tensor': merged_ids.to(device),
                    'input_mask': new_input_mask.to(device),
                    'target_mask': new_target_mask.to(device)}
        new_dict['decoder_attention_mask'] = torch.zeros_like(new_target_mask, dtype=torch.bool)

        mod_dict[target_mod] = new_dict
        return mod_dict
    
    def merge_sequences_batched(self, mod_dict, pred_ids, target_mod, text_tokenizer, default_sentinel="[S_1]"):
        # Unbatches and calls merge sequence per batch, then regroups it into a batch
        
        pad_id = text_tokenizer.token_to_id("[PAD]")
        
        B = mod_dict[target_mod]['tensor'].shape[0]
        device = mod_dict[target_mod]['tensor'].device

        tensors = torch.split(mod_dict[target_mod]['tensor'], 1)
        input_masks = torch.split(mod_dict[target_mod]['input_mask'], 1)
        pred_ids = torch.split(pred_ids, 1)

        input_dicts = []
        for t, im in zip(tensors, input_masks):
            d = {target_mod: {'tensor': t, 'input_mask': im}}
            input_dicts.append(d)

        merged_tensors = []
        merged_input_masks = []
        merged_target_masks = []
        merged_seq_lens = []
        for input_d, pi in zip(input_dicts, pred_ids):
            # Output of merge_sequences is mod_dict with modified target mod
            merged_d = self.merge_sequences(input_d, pi, target_mod, text_tokenizer, default_sentinel)[target_mod]
            merged_tensors.append(merged_d['tensor'])
            merged_input_masks.append(merged_d['input_mask'])
            merged_target_masks.append(merged_d['input_mask'])
            merged_seq_lens.append(merged_d['tensor'].shape[1])


        max_seq_len = max(merged_seq_lens)

        for i in range(len(merged_tensors)):
            # Right pad all tensors
            p1d = (0, max_seq_len - merged_seq_lens[i])
            merged_tensors[i] = F.pad(merged_tensors[i], p1d, "constant",pad_id)
            merged_input_masks[i] = F.pad(merged_input_masks[i], p1d, "constant", True)
            merged_target_masks[i] = F.pad(merged_target_masks[i], p1d, "constant", True)

        new_dict = {'tensor': torch.cat(merged_tensors, dim=0).to(device),
                    'input_mask': torch.cat(merged_input_masks, dim=0).to(device),
                    'target_mask': torch.cat(merged_target_masks, dim=0).to(device)}
        new_dict['decoder_attention_mask'] = torch.zeros_like(new_dict['target_mask'], dtype=torch.bool)

        mod_dict[target_mod] = new_dict
        return mod_dict

    def forward_enc_dec_maskgit_batched(self, mod_dict, target_mod, seed=None):
        # Encoder
        encoder_mod_dict = {mod: self.model.encoder_embeddings[mod](d)
                            for mod, d in mod_dict.items()
                            if mod in self.model.encoder_embeddings}
        encoder_tokens, encoder_emb, encoder_mask, encoder_mod_mask = self.forward_mask_encoder_generation(encoder_mod_dict)
        x = encoder_tokens + encoder_emb
        x = self.model.forward_encoder(x, encoder_mask)

        # Decoder
        context = self.model.decoder_proj_context(x) + encoder_emb
        decoder_mod_dict = {target_mod: self.model.decoder_embeddings[target_mod].forward_embed(mod_dict[target_mod])}
        decoder_tokens, decoder_emb, decoder_mask, decoder_mod_mask, mod_pos = self.forward_mask_decoder_maskgit(decoder_mod_dict, target_mod, seed=seed)
        y = decoder_tokens + decoder_emb
        y = self.model.forward_decoder(y, context, encoder_mask, None)
        B, N, D = y.shape
        logits = self.model.forward_logits(y, decoder_mod_dict, decoder_mod_mask)[target_mod]
        logits = logits.reshape(B, N, -1)


        return logits, mod_pos

    def maskgit_step_batched(self, mod_dict, target_mod, num_select, temperature, top_k, top_p, seed=None):        
        logits, mod_pos = self.forward_enc_dec_maskgit_batched(mod_dict, target_mod, seed=seed)

        # MaskGIT sampling
        top_samples, top_indices = self.select_tokens_batched(logits, num_select,
                                                      temperature=temperature, top_k=top_k, top_p=top_p)
        # Update mod dict
        # We rely on gather / scatter for batched operations
        top_pos = torch.gather(mod_pos, -1, top_indices) # (B, num_select)
        mod_dict[target_mod]['tensor'] = torch.scatter(mod_dict[target_mod]['tensor'], -1, top_pos, top_samples)
        mod_dict[target_mod]['input_mask'] = torch.scatter(mod_dict[target_mod]['input_mask'], -1, top_pos, torch.zeros_like(top_samples, dtype=torch.bool))
        mod_dict[target_mod]['target_mask'] = torch.scatter(mod_dict[target_mod]['target_mask'], -1, top_pos, torch.ones_like(top_samples, dtype=torch.bool))

        return mod_dict

    def guided_maskgit_step_batched(self, mod_dict, target_mod, num_select, temperature, top_k, top_p, 
                                    conditioning=[], guidance_scale=1.0, seed=None, write_all_predictions=False):

        ### 1 - First pass, with conditioning
        logits_cond, _ = self.forward_enc_dec_maskgit_batched(mod_dict, target_mod, seed=seed)

        ### 2 - Second pass, without conditioning
        mod_dict_uncond = copy.deepcopy(mod_dict)
        for mod in conditioning:
            if self.model.modality_info[mod]['type'] in ['seq', 'seq_token']:
                mod_dict_uncond = empty_seq_modality(mod_dict_uncond, mod)
            elif self.model.modality_info[mod]['type'] in ['seq_emb']: 
                mod_dict_uncond = empty_seq_emb_modality(mod_dict_uncond, mod)
            else:
                mod_dict_uncond = empty_img_modality(mod_dict_uncond, mod)

        logits_uncond, mod_pos = self.forward_enc_dec_maskgit_batched(mod_dict_uncond, target_mod, seed=seed)

        ### 3 - Classifier-free guidance
        logits = logits_uncond + (logits_cond - logits_uncond) * guidance_scale

        ### 4 - MaskGIT sampling
        top_samples, top_indices, all_samples = self.select_tokens_batched(
            logits, num_select,
            temperature=temperature, top_k=top_k, top_p=top_p, 
            return_all_samples=True
        )

        ### 5 - Update mod dict
        # We rely on gather / scatter for batched operations
        top_pos = torch.gather(mod_pos, -1, top_indices) # (B, num_select)
        if write_all_predictions:
            mod_dict[target_mod]['tensor'][:, mod_pos] = all_samples
        else:
            mod_dict[target_mod]['tensor'] = torch.scatter(mod_dict[target_mod]['tensor'], -1, top_pos, top_samples)
        mod_dict[target_mod]['input_mask'] = torch.scatter(mod_dict[target_mod]['input_mask'], -1, top_pos, torch.zeros_like(top_samples, dtype=torch.bool))
        mod_dict[target_mod]['target_mask'] = torch.scatter(mod_dict[target_mod]['target_mask'], -1, top_pos, torch.ones_like(top_samples, dtype=torch.bool))

        return mod_dict

    def multi_guided_maskgit_step_batched(self, uncond_dict, cond_dicts, cond_weights, target_mod, num_select, 
                                          temperature, top_k, top_p, seed=None, write_all_predictions=False):

        ### 1 - Conditional forward passes (one for each guided condition)
        logits_cond_all = []
        for cond_dict in cond_dicts:
            logits_cond_i, _ = self.forward_enc_dec_maskgit_batched(cond_dict, target_mod, seed=seed)
            logits_cond_all.append(logits_cond_i)
        
        ### 2 - Unconditional forward pass
        logits_uncond, mod_pos = self.forward_enc_dec_maskgit_batched(uncond_dict, target_mod, seed=seed)

        ### 3 Conjunction of multiple conditions: l_uncond + sum_i{w_i * (l_cond_i - l_uncond)}
        # See https://arxiv.org/abs/2206.01714
        logits = logits_uncond + torch.stack([w * (logits_cond - logits_uncond) for w, logits_cond in zip(cond_weights, logits_cond_all)]).sum(dim=0)

        ### 4 - MaskGIT sampling
        top_samples, top_indices, all_samples = self.select_tokens_batched(
            logits, num_select,
            temperature=temperature, top_k=top_k, top_p=top_p, 
            return_all_samples=True
        )
        
        ### 5 - Update mod dict with newly generated tokens
        # We rely on gather / scatter for batched operations
        top_pos = torch.gather(mod_pos, -1, top_indices) # (B, num_select)
        if write_all_predictions:
            uncond_dict[target_mod]['tensor'][:, mod_pos] = all_samples
        else:
            uncond_dict[target_mod]['tensor'] = torch.scatter(uncond_dict[target_mod]['tensor'], -1, top_pos, top_samples)
        uncond_dict[target_mod]['input_mask'] = torch.scatter(uncond_dict[target_mod]['input_mask'], -1, top_pos, torch.zeros_like(top_samples, dtype=torch.bool))
        uncond_dict[target_mod]['target_mask'] = torch.scatter(uncond_dict[target_mod]['target_mask'], -1, top_pos, torch.ones_like(top_samples, dtype=torch.bool))
        # Update conditioning dicts
        for i in range(len(cond_dicts)):
            cond_dicts[i][target_mod]['tensor'] = torch.scatter(cond_dicts[i][target_mod]['tensor'], -1, top_pos, top_samples)
            cond_dicts[i][target_mod]['input_mask'] = torch.scatter(cond_dicts[i][target_mod]['input_mask'], -1, top_pos, torch.zeros_like(top_samples, dtype=torch.bool))
            cond_dicts[i][target_mod]['target_mask'] = torch.scatter(cond_dicts[i][target_mod]['target_mask'], -1, top_pos, torch.ones_like(top_samples, dtype=torch.bool))

        return uncond_dict, cond_dicts

    def forward_enc_dec_roar_batched(self, mod_dict, target_mod, num_select, seed=None):
        # Encoder
        encoder_mod_dict = {mod: self.model.encoder_embeddings[mod](d)
                            for mod, d in mod_dict.items()
                            if mod in self.model.encoder_embeddings}
        encoder_tokens, encoder_emb, encoder_mask, encoder_mod_mask = self.forward_mask_encoder_generation(encoder_mod_dict)
        x = encoder_tokens + encoder_emb
        x = self.model.forward_encoder(x, encoder_mask)

        # Decoder
        context = self.model.decoder_proj_context(x) + encoder_emb
        decoder_mod_dict = {target_mod: self.model.decoder_embeddings[target_mod].forward_embed(mod_dict[target_mod])}
        decoder_tokens, decoder_emb, decoder_mask, decoder_mod_mask, mod_pos = self.forward_mask_decoder_roar(decoder_mod_dict, target_mod, num_select, seed=seed)
        y = decoder_tokens + decoder_emb
        y = self.model.forward_decoder(y, context, encoder_mask, None)
        B, N, D = y.shape
        logits = self.model.forward_logits(y, decoder_mod_dict, decoder_mod_mask)[target_mod]
        logits = logits.reshape(B, N, -1)

        return logits, mod_pos

    def roar_step_batched(self, mod_dict, target_mod, num_select, temperature, top_k, top_p, seed=None):
        """ROAR = Random Order Autoregression"""

        logits, mod_pos = self.forward_enc_dec_roar_batched(mod_dict, target_mod, num_select, seed=seed)

        # Simple sampling
        samples, sampled_probs = self.sample_tokens_batched(logits, temperature, top_k=top_k, top_p=top_p)

        # Update mod dict
        # We rely on scatter for batched operations
        select_pos = mod_pos
        mod_dict[target_mod]['tensor'] = torch.scatter(mod_dict[target_mod]['tensor'], -1, select_pos, samples)
        mod_dict[target_mod]['input_mask'] = torch.scatter(mod_dict[target_mod]['input_mask'], -1, select_pos, torch.zeros_like(samples, dtype=torch.bool))
        mod_dict[target_mod]['target_mask'] = torch.scatter(mod_dict[target_mod]['target_mask'], -1, select_pos, torch.ones_like(samples, dtype=torch.bool))

        return mod_dict

    def guided_roar_step_batched(self, mod_dict, target_mod, num_select, temperature, top_k, top_p, 
                                 conditioning=[], guidance_scale=1.0, seed=None):
        """ROAR = Random Order Autoregression"""

        ### 1 - First pass, with conditioning
        logits_cond, _ = self.forward_enc_dec_roar_batched(mod_dict, target_mod, num_select, seed=seed)

        ### 2 - Second pass, without conditioning
        mod_dict_uncond = copy.deepcopy(mod_dict)
        for mod in conditioning:
            if self.model.modality_info[mod]['type'] in ['seq', 'seq_token']:
                mod_dict_uncond = empty_seq_modality(mod_dict_uncond, mod)
            elif self.model.modality_info[mod]['type'] in ['seq_emb']: 
                mod_dict_uncond = empty_seq_emb_modality(mod_dict_uncond, mod)
            else:
                mod_dict_uncond = empty_img_modality(mod_dict_uncond, mod)

        logits_uncond, mod_pos = self.forward_enc_dec_roar_batched(mod_dict_uncond, target_mod, num_select, seed=seed)

        ### 3 - Classifier-free guidance
        logits = logits_uncond + (logits_cond - logits_uncond) * guidance_scale

        ### 4 - Simple sampling
        samples, sampled_probs = self.sample_tokens_batched(logits, temperature, top_k=top_k, top_p=top_p)

        ### 5 - Update mod dict
        # We rely on gather / scatter for batched operations
        select_pos = mod_pos
        mod_dict[target_mod]['tensor'] = torch.scatter(mod_dict[target_mod]['tensor'], -1, select_pos, samples)
        mod_dict[target_mod]['input_mask'] = torch.scatter(mod_dict[target_mod]['input_mask'], -1, select_pos, torch.zeros_like(samples, dtype=torch.bool))
        mod_dict[target_mod]['target_mask'] = torch.scatter(mod_dict[target_mod]['target_mask'], -1, select_pos, torch.ones_like(samples, dtype=torch.bool))

        return mod_dict

    def multi_guided_roar_step_batched(self, uncond_dict, cond_dicts, cond_weights, target_mod, 
                                       num_select, temperature, top_k, top_p, seed=None):

        ### 1 - Conditional forward passes (one for each guided condition)
        logits_cond_all = []
        for cond_dict in cond_dicts:
            logits_cond_i, _ = self.forward_enc_dec_roar_batched(cond_dict, target_mod, num_select, seed=seed)
            logits_cond_all.append(logits_cond_i)
        
        ### 2 - Unconditional forward pass
        logits_uncond, mod_pos = self.forward_enc_dec_roar_batched(uncond_dict, target_mod, num_select, seed=seed)

        ### 3 Conjunction of multiple conditions: l_uncond + sum_i{w_i * (l_cond_i - l_uncond)}
        # See https://arxiv.org/abs/2206.01714
        logits = logits_uncond + torch.stack([w * (logits_cond - logits_uncond) for w, logits_cond in zip(cond_weights, logits_cond_all)]).sum(dim=0)

        ### 4 - Simple sampling
        samples, sampled_probs = self.sample_tokens_batched(logits, temperature, top_k=top_k, top_p=top_p)
        
        ### 5 - Update mod dict
        # We rely on gather / scatter for batched operations
        select_pos = mod_pos
        uncond_dict[target_mod]['tensor'] = torch.scatter(uncond_dict[target_mod]['tensor'], -1, select_pos, samples)
        uncond_dict[target_mod]['input_mask'] = torch.scatter(uncond_dict[target_mod]['input_mask'], -1, select_pos, torch.zeros_like(samples, dtype=torch.bool))
        uncond_dict[target_mod]['target_mask'] = torch.scatter(uncond_dict[target_mod]['target_mask'], -1, select_pos, torch.ones_like(samples, dtype=torch.bool))
        # Update conditioning dicts
        for i in range(len(cond_dicts)):
            cond_dicts[i][target_mod]['tensor'] = torch.scatter(cond_dicts[i][target_mod]['tensor'], -1, select_pos, samples)
            cond_dicts[i][target_mod]['input_mask'] = torch.scatter(cond_dicts[i][target_mod]['input_mask'], -1, select_pos, torch.ones_like(samples, dtype=torch.bool))
            cond_dicts[i][target_mod]['target_mask'] = torch.scatter(cond_dicts[i][target_mod]['target_mask'], -1, select_pos, torch.zeros_like(samples, dtype=torch.bool))
        
        return uncond_dict, cond_dicts

    def autoregressive_step_batched(self, mod_dict, target_mod, temperature, top_k: float | int, top_p: float,
                                    use_eos=True, eos_token=None, start_tokens=None, text_tokenizer=None, seed=None):

        # Encoder
        encoder_mod_dict = {mod: self.model.encoder_embeddings[mod](d)
                            for mod, d in mod_dict.items()
                            if mod in self.model.encoder_embeddings}
        encoder_tokens, encoder_emb, encoder_mask, encoder_mod_mask = self.forward_mask_encoder_generation(encoder_mod_dict)
        x = encoder_tokens + encoder_emb
        x = self.model.forward_encoder(x, encoder_mask) # B, N, D

        # Get batch size
        B = x.shape[0]

        # Decoder
        context = self.model.decoder_proj_context(x) + encoder_emb
        decoder_mod_dict = {target_mod: self.model.decoder_embeddings[target_mod].forward_embed(mod_dict[target_mod])}
        decoder_ids, decoder_emb, decoder_mask, decoder_mod_mask, mod_pos = self.forward_mask_decoder_autoregressive(decoder_mod_dict, target_mod, seed=seed)
        device = decoder_ids.device
        seq_len = self.model.modality_info[target_mod]['max_tokens']

        if use_eos and eos_token is None:
            # The eos_token is the final sentinel token provided
            eos_token = decoder_ids[0][decoder_mask[0] == 0][-1] # Assumes the EOS token is the same for all
        if use_eos:
            eos_token = eos_token.to(device)

        # If no start_tokens, just use the beginning of the actual target (i.e., a sentinel token)
        out = decoder_ids[:, :1] if start_tokens is None else start_tokens.to(device)
        # Set decoder_tokens to None, we do not use them for decoding
        decoder_ids = None

        # If all samples of the batch have eos, return early
        if use_eos and (out == eos_token).any(dim=-1).all():
            return out

        y_emb = decoder_emb[:, :seq_len]
        seq_len = y_emb.shape[1]

        # Auto-regressive decoding and sampling
        for i in range(seq_len):
            cur_len = out.shape[1]
            # Convert ids into word embeddings and add corresponding posembs + modemb
            y = self.model.decoder_embeddings[target_mod].token_emb(out) + y_emb[:, :cur_len]
            # Build causal mask
            causal_mask = torch.ones((cur_len, cur_len), dtype=torch.bool, device=y.device).triu(1)
            causal_mask = repeat(causal_mask, "n1 n2 -> b n1 n2", b=B)
            
            y = self.model.forward_decoder(y, context, encoder_mask, causal_mask)
            logits = self.model.forward_logits(y, decoder_mod_dict, decoder_mod_mask[:, :cur_len])[target_mod]
            logits = rearrange(logits, "(b n) d -> b n d", b=B, n=cur_len)
            last_logits = logits[:, -1]

            # Sample token for the newly generated logit
            if np.isclose(temperature, 0, atol=1e-10):
                sample = torch.argmax(last_logits, dim=-1, keepdim=True)
            else:
                filtered_logits = self.top_k_top_p_filtering(last_logits, top_k, top_p)
                probs = F.softmax(filtered_logits / temperature, dim=-1)
                sample = torch.multinomial(probs, 1)
            out = torch.cat((out, sample), dim=-1)

            if use_eos and (out == eos_token).any(dim=-1).all():
                break

        mod_dict = self.merge_sequences_batched(mod_dict, out, target_mod, text_tokenizer)

        return mod_dict

    def guided_autoregressive_step_batched(self, mod_dict, target_mod, temperature, top_k: float | int, top_p: float,
                                           use_eos=True, eos_token=None, start_tokens=None, text_tokenizer=None, 
                                           conditioning=[], guidance_scale=1.0, seed=None):

         ### 1 - Encoder forward pass, with conditioning
        
        # Encoder
        encoder_mod_dict = {mod: self.model.encoder_embeddings[mod](d)
                            for mod, d in mod_dict.items()
                            if mod in self.model.encoder_embeddings}
        encoder_tokens, encoder_emb, encoder_mask_cond, encoder_mod_mask = self.forward_mask_encoder_generation(encoder_mod_dict)
        x = encoder_tokens + encoder_emb
        x = self.model.forward_encoder(x, encoder_mask_cond) # B, N, D

        # Get batch size
        B = x.shape[0]

        # Decoder
        context_cond = self.model.decoder_proj_context(x) + encoder_emb
        decoder_mod_dict_cond = {target_mod: self.model.decoder_embeddings[target_mod].forward_embed(mod_dict[target_mod])}
        decoder_ids, decoder_emb, decoder_mask, decoder_mod_mask_cond, mod_pos = self.forward_mask_decoder_autoregressive(decoder_mod_dict_cond, target_mod, seed=seed)
        device = decoder_ids.device
        seq_len = self.model.modality_info[target_mod]['max_tokens']


        ### 2 - Encoder forward pass, without conditioning

        mod_dict_uncond = copy.deepcopy(mod_dict)
        for mod in conditioning:
            if self.model.modality_info[mod]['type'] in ['seq', 'seq_token']:
                mod_dict_uncond = empty_seq_modality(mod_dict_uncond, mod)
            elif self.model.modality_info[mod]['type'] in ['seq_emb']: 
                mod_dict_uncond = empty_seq_emb_modality(mod_dict_uncond, mod)
            else:
                mod_dict_uncond = empty_img_modality(mod_dict_uncond, mod)

        # Encoder
        encoder_mod_dict = {mod: self.model.encoder_embeddings[mod](d)
                            for mod, d in mod_dict_uncond.items()
                            if mod in self.model.encoder_embeddings}
        encoder_tokens, encoder_emb, encoder_mask_uncond, encoder_mod_mask = self.forward_mask_encoder_generation(encoder_mod_dict)
        x = encoder_tokens + encoder_emb
        x = self.model.forward_encoder(x, encoder_mask_uncond) # B, N, D

        # Decoder
        context_uncond = self.model.decoder_proj_context(x) + encoder_emb
        decoder_mod_dict_uncond = {target_mod: self.model.decoder_embeddings[target_mod].forward_embed(mod_dict[target_mod])}
        decoder_ids, decoder_emb, decoder_mask, decoder_mod_mask_uncond, mod_pos = self.forward_mask_decoder_autoregressive(decoder_mod_dict_uncond, target_mod, seed=seed)


        if use_eos and eos_token is None:
            # The eos_token is the final sentinel token provided
            eos_token = decoder_ids[0][decoder_mask[0] == 0][-1] # Assumes the EOS token is the same for all
        if use_eos:
            eos_token = eos_token.to(device)

        # If no start_tokens, just use the beginning of the actual target (i.e., a sentinel token)
        out = decoder_ids[:, :1] if start_tokens is None else start_tokens.to(device)
        # Set decoder_tokens to None, we do not use them for decoding
        decoder_ids = None

        # If all samples of the batch have eos, return early
        if use_eos and (out == eos_token).any(dim=-1).all():
            return out

        y_emb = decoder_emb[:, :seq_len]
        seq_len = y_emb.shape[1]

        ### 3 -  Auto-regressive decoding and sampling
        for i in range(seq_len):
            cur_len = out.shape[1]
            # Convert ids into word embeddings and add corresponding posembs + modemb
            y = self.model.decoder_embeddings[target_mod].token_emb(out) + y_emb[:, :cur_len]
            # Build causal mask
            causal_mask = torch.ones((cur_len, cur_len), dtype=torch.bool, device=y.device).triu(1)
            causal_mask = repeat(causal_mask, "n1 n2 -> b n1 n2", b=B)
            
            ### 3a - Decoder forward pass, with conditioning
            y_cond = self.model.forward_decoder(y, context_cond, encoder_mask_cond, causal_mask)
            logits_cond = self.model.forward_logits(y_cond, decoder_mod_dict_cond, decoder_mod_mask_cond[:, :cur_len])[target_mod]
            logits_cond = rearrange(logits_cond, "(b n) d -> b n d", b=B, n=cur_len)
            last_logits_cond = logits_cond[:, -1]

            ### 3b - Decoder forward pass, without conditioning
            y_uncond = self.model.forward_decoder(y, context_uncond, encoder_mask_uncond, causal_mask)
            logits_uncond = self.model.forward_logits(y_uncond, decoder_mod_dict_uncond, decoder_mod_mask_uncond[:, :cur_len])[target_mod]
            logits_uncond = rearrange(logits_uncond, "(b n) d -> b n d", b=B, n=cur_len)
            last_logits_uncond = logits_uncond[:, -1]

            ### 3c - Classifier-free guidance
            last_logits = last_logits_uncond + (last_logits_cond - last_logits_uncond) * guidance_scale

            # Sample token for the newly generated logit
            if np.isclose(temperature, 0, atol=1e-10):
                sample = torch.argmax(last_logits, dim=-1, keepdim=True)
            else:
                filtered_logits = self.top_k_top_p_filtering(last_logits, top_k, top_p)
                probs = F.softmax(filtered_logits / temperature, dim=-1)
                sample = torch.multinomial(probs, 1)
            out = torch.cat((out, sample), dim=-1)

            if use_eos and (out == eos_token).any(dim=-1).all():
                break

        mod_dict = self.merge_sequences_batched(mod_dict, out, target_mod, text_tokenizer)

        return mod_dict


    @torch.no_grad()
    def generate(self, mod_dict, schedule, top_k=0.0, top_p=0.0, tokenizer=None, verbose=False, seed=None,
                 num_tokens=None):
        """ Generates a sequence of tokens from the input modalities.
        :param mod_dict: Dictionary of modalities.
        :param schedule: Schedule of modalities to use. 
            list of dictionaries containing {target_domain, scheme, num_tokens, temperature, cfg_scale, cfg_cond_domains}.
        :param top_k: top_k > 0: Keep only top k tokens with highest probability (a.k.a. top-k filtering).
        :param top_p: top_p > 0.0: Keep the top tokens with cumulative probability >= top_p (a.k.a. nucleus filtering).
        :param tokenizer: Modality tokenizers: dict[domain, module] with each module having an text_tokenizer attribute
        :param verbose: Whether to print progress.
        :param seed: Random seed.
        :return: Generated mod dict.
        """

        # Input embedding -> tokenizes the modalities - Many are placeholder for now
        mod_dict = copy.deepcopy(mod_dict)

        for step, schedule_step_info in tqdm(enumerate(schedule), disable=not verbose):
            target_mod = schedule_step_info['target_domain']
            temp = schedule_step_info['temperature']
            cfg_scale = schedule_step_info.get('cfg_scale', 1.0)
            cfg_conditioning = schedule_step_info.get('cfg_cond_domains', [])
            seed_i = seed + step if seed is not None else None
            
            if self.model.modality_info[target_mod]['type'] == 'img':
                scheme = schedule_step_info['scheme']
                num_select = num_tokens or schedule_step_info['num_tokens']

                if scheme.lower() == 'maskgit':
                    if cfg_scale == 1.0 or len(cfg_conditioning) == 0:
                        mod_dict = self.maskgit_step_batched(
                            mod_dict, target_mod, num_select, temperature=temp, 
                            top_k=top_k, top_p=top_p, seed=seed_i
                        )
                    else:
                        mod_dict = self.guided_maskgit_step_batched(
                            mod_dict, target_mod, num_select, temperature=temp, top_k=top_k, top_p=top_p,
                            conditioning=cfg_conditioning, guidance_scale=cfg_scale, seed=seed_i
                        )
                elif scheme.lower() == 'roar':
                    if cfg_scale == 1.0 or len(cfg_conditioning) == 0:
                        mod_dict = self.roar_step_batched(
                            mod_dict, target_mod, num_select, temperature=temp, 
                            top_k=top_k, top_p=top_p, seed=seed_i
                        )
                    else:
                        mod_dict = self.guided_roar_step_batched(
                            mod_dict, target_mod, num_select, temperature=temp, top_k=top_k, top_p=top_p, 
                            conditioning=cfg_conditioning, guidance_scale=cfg_scale, seed=seed_i
                        )
                else:
                    raise ValueError("Invalid sampling scheme")
            elif self.model.modality_info[target_mod]['type'] in ['seq', 'seq_token']:
                if cfg_scale == 1.0 or len(cfg_conditioning) == 0:
                    mod_dict = self.autoregressive_step_batched(
                        mod_dict, target_mod, temperature=temp, top_k=top_k, top_p=top_p,
                        text_tokenizer=tokenizer[target_mod].text_tokenizer, seed=seed_i
                    )
                else:
                    mod_dict = self.guided_autoregressive_step_batched(
                        mod_dict, target_mod, temperature=temp, top_k=top_k, top_p=top_p,
                        text_tokenizer=tokenizer[target_mod].text_tokenizer, conditioning=cfg_conditioning,
                        guidance_scale=cfg_scale, seed=seed_i
                    )
            else:
                raise ValueError("Invalid schedule")

        return mod_dict


    @torch.no_grad()
    def generate_iter(self, mod_dict, schedule, top_k=0.0, top_p=0.0, text_tokenizer=None, verbose=False, seed=None):
        """ Iterator that generates a sequence of tokens from the input modalities step by step.
        :param mod_dict: Dictionary of modalities.
        :param schedule: Schedule of modalities to use. 
            list of dictionaries containing {target_domain, scheme, num_tokens, temperature, cfg_scale, cfg_cond_domains}.
        :param top_k: top_k > 0: Keep only top k tokens with highest probability (a.k.a. top-k filtering).
        :param top_p: top_p > 0.0: Keep the top tokens with cumulative probability >= top_p (a.k.a. nucleus filtering).
        :param text_tokenizer: Text tokenizer.
        :param verbose: Whether to print progress.
        :param seed: Random seed.
        :return: Iterator of generated mod dict.
        """

        # Input embedding -> tokenizes the modalities - Many are placeholder for now
        mod_dict = copy.deepcopy(mod_dict)

        for step, schedule_step_info in tqdm(enumerate(schedule), disable=not verbose):
            target_mod = schedule_step_info['target_domain']
            temp = schedule_step_info['temperature']
            cfg_scale = schedule_step_info.get('cfg_scale', 1.0)
            cfg_conditioning = schedule_step_info.get('cfg_cond_domains', [])
            seed_i = seed + step if seed is not None else None
            
            if self.model.modality_info[target_mod]['type'] == 'img':
                scheme = schedule_step_info['scheme']
                num_select = schedule_step_info['num_tokens']    

                if scheme.lower() == 'maskgit':
                    if cfg_scale == 1.0 or len(cfg_conditioning) == 0:
                        mod_dict = self.maskgit_step_batched(
                            mod_dict, target_mod, num_select, temperature=temp, 
                            top_k=top_k, top_p=top_p, seed=seed_i
                        )
                    else:
                        mod_dict = self.guided_maskgit_step_batched(
                            mod_dict, target_mod, num_select, temperature=temp, top_k=top_k, top_p=top_p,
                            conditioning=cfg_conditioning, guidance_scale=cfg_scale, seed=seed_i,
                            write_all_predictions=True
                        )
                elif scheme.lower() == 'roar':
                    if cfg_scale == 1.0 or len(cfg_conditioning) == 0:
                        mod_dict = self.roar_step_batched(
                            mod_dict, target_mod, num_select, temperature=temp, 
                            top_k=top_k, top_p=top_p, seed=seed_i
                        )
                    else:
                        mod_dict = self.guided_roar_step_batched(
                            mod_dict, target_mod, num_select, temperature=temp, top_k=top_k, top_p=top_p, 
                            conditioning=cfg_conditioning, guidance_scale=cfg_scale, seed=seed_i
                        )
                else:
                    raise ValueError("Invalid sampling scheme")
            elif self.model.modality_info[target_mod]['type'] in ['seq', 'seq_token']:
                if cfg_scale == 1.0 or len(cfg_conditioning) == 0:
                    mod_dict = self.autoregressive_step_batched(
                        mod_dict, target_mod, temperature=temp, top_k=top_k, top_p=top_p,
                        text_tokenizer=text_tokenizer, seed=seed_i
                    )
                else:
                    mod_dict = self.guided_autoregressive_step_batched(
                        mod_dict, target_mod, temperature=temp, top_k=top_k, top_p=top_p,
                        text_tokenizer=text_tokenizer, conditioning=cfg_conditioning, 
                        guidance_scale=cfg_scale, seed=seed_i
                    )
            else:
                raise ValueError("Invalid schedule")

            yield mod_dict

    @torch.no_grad()
    def generate_multi_guided(self, uncond_dict, cond_dicts, schedule, top_k=0.0, top_p=0.0, 
                              text_tokenizer=None, verbose=False, seed=None):
        # Generation function for multiple weighted conditions

        # To detect when a modality has finished generating, we keep track of the current target modality
        cur_target_mod = schedule[0]['target_domain']

        uncond_dict = copy.deepcopy(uncond_dict)
        cond_dicts = copy.deepcopy(cond_dicts)

        # Add the to-be-generated modality to the conditional dicts
        for i in range(len(cond_dicts)):
            cond_dicts[i][cur_target_mod] = copy.deepcopy(uncond_dict[cur_target_mod])

        for step, schedule_step_info in tqdm(enumerate(schedule), disable=not verbose):
            target_mod = schedule_step_info['target_domain']
            temp = schedule_step_info['temperature']
            num_select = schedule_step_info['num_tokens']
            cond_weights = schedule_step_info['cfg_scale']

            # Once a modality is fully generated, add it as a new condition
            if cur_target_mod != target_mod:
                for i in range(len(cond_dicts)):
                    # Remove the previously generated modality from the conditionings
                    del cond_dicts[i][cur_target_mod]
                    # Add the next modality to be generated to the conditionings
                    cond_dicts[i][target_mod] = copy.deepcopy(uncond_dict[target_mod])
                
                # Remove the fully generated modality from the unconditional dict inputs
                uncond_dict[cur_target_mod]['input_mask'][:] = True

                # Add the previously generated modality as an additional condition
                new_cond = {}
                new_cond[cur_target_mod] = copy.deepcopy(uncond_dict[cur_target_mod])
                new_cond[cur_target_mod]['input_mask'][:] = False
                new_cond[cur_target_mod]['target_mask'][:] = True
                new_cond[target_mod] = copy.deepcopy(uncond_dict[target_mod])
                cond_dicts.append(new_cond)

                cur_target_mod = target_mod

            if self.model.modality_info[target_mod]['type'] == 'img':
                scheme = schedule_step_info['scheme']

                if scheme.lower() == 'maskgit':
                    uncond_dict, cond_dicts = self.multi_guided_maskgit_step_batched(
                        uncond_dict, cond_dicts, cond_weights, target_mod, num_select, temp, top_k, top_p, seed=seed
                    )
                elif scheme.lower() == 'roar':
                    uncond_dict, cond_dicts = self.multi_guided_roar_step_batched(
                        uncond_dict, cond_dicts, cond_weights, target_mod, num_select, temp, top_k, top_p, seed=seed
                    )
                else:
                    raise ValueError("Invalid sampling scheme")

            else:
                raise NotImplementedError("Only image modalities are supported for now")

        return uncond_dict


# Copyright (c) OpenMMLab. All rights reserved.
import warnings
from typing import Dict

import torch
from torch.nn import Upsample as Upsample
import torch.nn as nn
import torch.utils.checkpoint as cp

from torch.nn.modules.batchnorm import _BatchNorm
from torch.nn.modules.instancenorm import _InstanceNorm
SyncBatchNorm_ = nn.SyncBatchNorm

from terratorch.models.backbones.utils import UpConvBlock, BasicConvBlock
from terratorch.models.decoders.utils import ConvModule

from terratorch.registry import TERRATORCH_BACKBONE_REGISTRY

@TERRATORCH_BACKBONE_REGISTRY.register
class UNet(nn.Module):
    """UNet backbone.

    This backbone is the implementation of `U-Net: Convolutional Networks
    for Biomedical Image Segmentation <https://arxiv.org/abs/1505.04597>`_.

    Args:
        in_channels (int): Number of input image channels. Default" 3.
        out_channels (int): Number of base channels of each stage.
            The output channels of the first stage. Default: 64.
        num_stages (int): Number of stages in encoder, normally 5. Default: 5.
        strides (Sequence[int 1 | 2]): Strides of each stage in encoder.
            len(strides) is equal to num_stages. Normally the stride of the
            first stage in encoder is 1. If strides[i]=2, it uses stride
            convolution to downsample in the correspondence encoder stage.
            Default: (1, 1, 1, 1, 1).
        enc_num_convs (Sequence[int]): Number of convolutional layers in the
            convolution block of the correspondence encoder stage.
            Default: (2, 2, 2, 2, 2).
        dec_num_convs (Sequence[int]): Number of convolutional layers in the
            convolution block of the correspondence decoder stage.
            Default: (2, 2, 2, 2).
        downsamples (Sequence[int]): Whether use MaxPool to downsample the
            feature map after the first stage of encoder
            (stages: [1, num_stages)). If the correspondence encoder stage use
            stride convolution (strides[i]=2), it will never use MaxPool to
            downsample, even downsamples[i-1]=True.
            Default: (True, True, True, True).
        enc_dilations (Sequence[int]): Dilation rate of each stage in encoder.
            Default: (1, 1, 1, 1, 1).
        dec_dilations (Sequence[int]): Dilation rate of each stage in decoder.
            Default: (1, 1, 1, 1).
        with_cp (bool): Use checkpoint or not. Using checkpoint will save some
            memory while slowing down the training speed. Default: False.
        conv_cfg (dict | None): Config dict for convolution layer.
            Default: None.
        norm_cfg (dict | None): Config dict for normalization layer.
            Default: dict(type='BN').
        act_cfg (dict | None): Config dict for activation layer in ConvModule.
            Default: dict(type='ReLU').
        upsample_cfg (dict): The upsample config of the upsample module in
            decoder. Default: dict(type='InterpConv').
        norm_eval (bool): Whether to set norm layers to eval mode, namely,
            freeze running stats (mean and var). Note: Effect on Batch Norm
            and its variants only. Default: False.
        dcn (bool): Use deformable convolution in convolutional layer or not.
            Default: None.
        plugins (dict): plugins for convolutional layers. Default: None.
        pretrained (str, optional): model pretrained path. Default: None
        init_cfg (dict or list[dict], optional): Initialization config dict.
            Default: None

    Notice:
        The input image size should be divisible by the whole downsample rate
        of the encoder. More detail of the whole downsample rate can be found
        in `UNet._check_input_divisible`.
    """

    def __init__(self,
                 in_channels=3,
                 out_channels=64,
                 num_stages=5,
                 strides=(1, 1, 1, 1, 1),
                 enc_num_convs=(2, 2, 2, 2, 2),
                 dec_num_convs=(2, 2, 2, 2),
                 downsamples=(True, True, True, True),
                 enc_dilations=(1, 1, 1, 1, 1),
                 dec_dilations=(1, 1, 1, 1),
                 with_cp=False,
                 conv_cfg=None,
                 norm_cfg=dict(type='BN'),
                 act_cfg=dict(type='ReLU'),
                 upsample_cfg=None,
                 norm_eval=False,
                 dcn=None,
                 plugins=None,
                 pretrained=None,
                 init_cfg=None):
        super(UNet, self).__init__()

        self.pretrained = pretrained
        assert not (init_cfg and pretrained), \
            'init_cfg and pretrained cannot be setting at the same time'
        if isinstance(pretrained, str):
            warnings.warn('DeprecationWarning: pretrained is a deprecated, '
                          'please use "init_cfg" instead')
            self.init_cfg = dict(type='Pretrained', checkpoint=pretrained)
        elif pretrained is None:
            if init_cfg is None:
                self.init_cfg = [
                    dict(type='Kaiming', layer='Conv2d'),
                    dict(
                        type='Constant',
                        val=1,
                        layer=['_BatchNorm', 'GroupNorm'])
                ]
        else:
            raise TypeError('pretrained must be a str or None')

        assert dcn is None, 'Not implemented yet.'
        assert plugins is None, 'Not implemented yet.'
        assert len(strides) == num_stages, \
            'The length of strides should be equal to num_stages, '\
            f'while the strides is {strides}, the length of '\
            f'strides is {len(strides)}, and the num_stages is '\
            f'{num_stages}.'
        assert len(enc_num_convs) == num_stages, \
            'The length of enc_num_convs should be equal to num_stages, '\
            f'while the enc_num_convs is {enc_num_convs}, the length of '\
            f'enc_num_convs is {len(enc_num_convs)}, and the num_stages is '\
            f'{num_stages}.'
        assert len(dec_num_convs) == (num_stages-1), \
            'The length of dec_num_convs should be equal to (num_stages-1), '\
            f'while the dec_num_convs is {dec_num_convs}, the length of '\
            f'dec_num_convs is {len(dec_num_convs)}, and the num_stages is '\
            f'{num_stages}.'
        assert len(downsamples) == (num_stages-1), \
            'The length of downsamples should be equal to (num_stages-1), '\
            f'while the downsamples is {downsamples}, the length of '\
            f'downsamples is {len(downsamples)}, and the num_stages is '\
            f'{num_stages}.'
        assert len(enc_dilations) == num_stages, \
            'The length of enc_dilations should be equal to num_stages, '\
            f'while the enc_dilations is {enc_dilations}, the length of '\
            f'enc_dilations is {len(enc_dilations)}, and the num_stages is '\
            f'{num_stages}.'
        assert len(dec_dilations) == (num_stages-1), \
            'The length of dec_dilations should be equal to (num_stages-1), '\
            f'while the dec_dilations is {dec_dilations}, the length of '\
            f'dec_dilations is {len(dec_dilations)}, and the num_stages is '\
            f'{num_stages}.'
        self.num_stages = num_stages
        self.strides = strides
        self.downsamples = downsamples
        self.norm_eval = norm_eval
        self.out_channels = [out_channels * 2**i for i in reversed(range(num_stages))]

        self.encoder = nn.ModuleList()
        self.decoder = nn.ModuleList()

        for i in range(num_stages):
            enc_conv_block = []
            if i != 0:
                if strides[i] == 1 and downsamples[i - 1]:
                    enc_conv_block.append(nn.MaxPool2d(kernel_size=2))
                upsample = (strides[i] != 1 or downsamples[i - 1])
                self.decoder.append(
                    UpConvBlock(
                        conv_block=BasicConvBlock,
                        in_channels=out_channels * 2**i,
                        skip_channels=out_channels * 2**(i - 1),
                        out_channels=out_channels * 2**(i - 1),
                        num_convs=dec_num_convs[i - 1],
                        stride=1,
                        dilation=dec_dilations[i - 1],
                        with_cp=with_cp,
                        conv_cfg=conv_cfg,
                        norm_cfg=norm_cfg,
                        act_cfg=act_cfg,
                        upsample_cfg=upsample_cfg if upsample else None,
                        dcn=None,
                        plugins=None))

            enc_conv_block.append(
                BasicConvBlock(
                    in_channels=in_channels,
                    out_channels=out_channels * 2**i,
                    num_convs=enc_num_convs[i],
                    stride=strides[i],
                    dilation=enc_dilations[i],
                    with_cp=with_cp,
                    conv_cfg=conv_cfg,
                    norm_cfg=norm_cfg,
                    act_cfg=act_cfg,
                    dcn=None,
                    plugins=None))
            self.encoder.append((nn.Sequential(*enc_conv_block)))
            in_channels = out_channels * 2**i

    def forward(self, x):

        # We can check just the first image, since the batch 
        # already was approved by the stackability test, which means
        # all images has the same dimensions. 
        self._check_input_divisible(x[0])

        enc_outs = []
        for enc in self.encoder:
            x = enc(x)
            enc_outs.append(x)
        dec_outs = [x]
        for i in reversed(range(len(self.decoder))):
            x = self.decoder[i](enc_outs[i], x)
            dec_outs.append(x)
        return dec_outs

    def train(self, mode=True):
        """Convert the model into training mode while keep normalization layer
        freezed."""
        super(UNet, self).train(mode)
        if mode and self.norm_eval:
            for m in self.modules():
                # trick: eval have effect on BatchNorm only
                if isinstance(m, _BatchNorm):
                    m.eval()

    def _check_input_divisible(self, x):
        h, w = x.shape[-2:]
        whole_downsample_rate = 1
        for i in range(1, self.num_stages):
            if self.strides[i] == 2 or self.downsamples[i - 1]:
                whole_downsample_rate *= 2
        assert (h % whole_downsample_rate == 0) \
            and (w % whole_downsample_rate == 0),\
            f'The input image size {(h, w)} should be divisible by the whole '\
            f'downsample rate {whole_downsample_rate}, when num_stages is '\
            f'{self.num_stages}, strides is {self.strides}, and downsamples '\
            f'is {self.downsamples}.'



import importlib
import sys
from collections.abc import Callable
import logging 

import timm
import torch
from torch import nn
import logging

import terratorch.models.decoders as decoder_registry
from terratorch.models.backbones.clay_v1.embedder import Embedder
from terratorch.models.model import (
    AuxiliaryHead,
    Model,
    ModelFactory,
)
from terratorch.models.pixel_wise_model import PixelWiseModel
from terratorch.models.scalar_output_model import ScalarOutputModel
from terratorch.models.utils import DecoderNotFoundError, extract_prefix_keys
from terratorch.registry import MODEL_FACTORY_REGISTRY
from box import Box
from terratorch.models.backbones.clay_v15.model import ClayMAE

PIXEL_WISE_TASKS = ["segmentation", "regression"]
SCALAR_TASKS = ["classification", "scalar_regression"]
SUPPORTED_TASKS = PIXEL_WISE_TASKS + SCALAR_TASKS


class DecoderNotFoundError(Exception):
    pass

class ModelWrapper(nn.Module):

    def __init__(self, batch_size, bands, platform, model: nn.Module = None) -> None:

        super(ModelWrapper, self).__init__()
        self.batch_size = batch_size
        self.bands = bands
        self.platform = platform
        self.model = model

        #self.embedding_shape = self.model.state_dict()['decoder.embed_to_pixels.dem.bias'].shape[0]

    def channels(self):
        return (1, self.embedding_shape)

    @property
    def parameters(self):
        return self.model.parameters

    def forward(self, args, **kwargs):
        datacube = {}
        datacube['pixels'] = args
        datacube['time'] = torch.zeros(self.batch_size, 4)
        datacube['platform'] = self.platform
        datacube['latlon'] = torch.zeros(self.batch_size, 4)
        datacube['waves'] = torch.zeros(4)
        datacube['gsd'] = torch.tensor(10)
        return self.model.forward(datacube, **kwargs)

@MODEL_FACTORY_REGISTRY.register
class Clay1_5ModelFactory(ModelFactory):
    def build_model(
        self,
        task: str,
        backbone: str | nn.Module,
        decoder: str | nn.Module,
        in_channels: int,
        bands: list[int] = [],
        num_classes: int | None = None,
        pretrained: bool = True,  # noqa: FBT001, FBT002
        num_frames: int = 1,
        prepare_features_for_image_model: Callable | None = None,
        aux_decoders: list[AuxiliaryHead] | None = None,
        rescale: bool = True,  # noqa: FBT002, FBT001
        checkpoint_path: str = None,
        **kwargs,
    ) -> Model:
        # try:
        #     from claymodel.model import ClayMAE
        # except ImportError:
        #     message = "clay v1.5 not installed, please use pip install claimodel"
        #     logging.getLogger("terratorch").debug(message)
        #    raise Exception(message)
        batch_size = kwargs.get("batch_size")
        platform = kwargs.get("platform")
        kwargs["metadata"] = Box(kwargs["metadata"])
        return ModelWrapper(batch_size, bands, platform, ClayMAE(**kwargs))
